(* ============================================================ *)
(*  LightningMaker is the concrete class for emitting COM2+	*)
(*  intermediate language for the VOS.				*)
(*  Copyright (c) John Gough 1999, 2000.			*)
(* ============================================================ *)

MODULE LightningMaker;

  IMPORT 
	GPCPcopyright,
	Error,
	GPText,
	Console,
	LitValue,
	CPascalS,
	FileNames,
	CompState,
	LightningAsm,
	LightningBase,
	ClassMaker,
	GPTextFiles,
	Asm := IlasmCodes,
	Vos := IlasmUtil,
	G   := Builtin,
	D   := Symbols,
	Id  := IdDesc ,
	Ty  := TypeDesc,
	Xp  := ExprDesc,
	St  := StatDesc;

(* ============================================================ *)

  CONST pubStat     = Asm.att_public + Asm.att_static;
  	modAttrib   = Asm.att_public + Asm.att_sealed;
  	staticAtt   = Asm.att_static;
  	extern      = Asm.att_extern;

  CONST inlineLimit = 4; (* limit for inline expansion of element copies *)

(* ============================================================ *)

  TYPE LightningEmitter* = 	
	POINTER TO 
	  RECORD (LightningBase.ClassEmitter)
	 (* --------------------------- *
	  * mod* : Id.BlkId; 		*
	  * --------------------------- *)
	    work : D.IdSeq;
	    outF : Vos.IlasmFile;
	  END;

(* ------------------------------------ *)

  TYPE LightningAssembler* = 	
	POINTER TO 
	  RECORD (ClassMaker.Assembler)
	    emit : Vos.IlasmFile;
	  END;

(* ------------------------------------ *)

  VAR  asmName   : LitValue.CharOpen;
       asmExe    : BOOLEAN;

(* ============================================================ *)
(* ============================================================ *)

  PROCEDURE newLightningEmitter*(mod : Id.BlkId) : LightningEmitter;
    VAR emitter : LightningEmitter;
  BEGIN
    NEW(emitter);
    emitter.mod := mod;
    LightningBase.emitter := emitter;
    D.InitIdSeq(emitter.work, 4);
    D.AppendIdnt(emitter.work, mod);
    RETURN emitter;
  END newLightningEmitter;

(* ============================================================ *)

  PROCEDURE newLightningAsm*() : LightningAssembler;
    VAR asm : LightningAssembler;
  BEGIN
    NEW(asm);
    LightningAsm.Init();
    RETURN asm;
  END newLightningAsm;

(* ============================================================ *)
(*  Creates basic imports for System, and inserts a few type	*)
(*  descriptors for Object, Exception, and String.              *)
(* ============================================================ *)

  PROCEDURE (this : LightningEmitter)Init*();
    VAR tId : Id.TypId;
        blk : Id.BlkId;
        obj : Id.TypId;
        str : Id.TypId;
        exc : Id.TypId;
  BEGIN
   (*
    *  Create import descriptor for [mscorlib]System
    *)
    Builtin.MkDummyImport("mscorlib_System", "[mscorlib]System", blk); 
   (*
    *  Create various classes.
    *)
    Builtin.MkDummyClass("Object", blk, Ty.isAbs, obj);
    CompState.ntvObj := obj.type;
    Builtin.MkDummyClass("String", blk, Ty.noAtt, str);
    Builtin.SetPtrBase(str, obj);
    CompState.ntvStr := str.type;
    Builtin.MkDummyClass("Exception", blk, Ty.extns, exc);
    Builtin.SetPtrBase(exc, obj);
    CompState.ntvExc := exc.type;
   (*
    *  Create import descriptor for [RTS]RTS
    *)
    Builtin.MkDummyImport("RTS", "", blk);
    Builtin.MkDummyAlias("NativeObject", blk, obj.type);
    Builtin.MkDummyAlias("NativeString", blk, str.type);
    Builtin.MkDummyAlias("NativeException", blk, exc.type);
  END Init;

(* ============================================================ *)

  PROCEDURE (this : LightningAssembler)Assemble*();
  (** Overrides EMPTY method in ClassMaker *)
  BEGIN
    LightningAsm.Assemble(asmName, asmExe);
  END Assemble;

(* ============================================================ *)

  PROCEDURE (this : LightningEmitter)AddNewRecEmitter*(inTp : Ty.Record);
  (* Overrides AddNewRecEmitter() in LightningBase. *)
    VAR idnt : D.Idnt;
  BEGIN
    idnt := NIL;
    IF inTp.bindTp # NIL THEN idnt := inTp.bindTp.idnt;
    ELSIF inTp.idnt # NIL THEN idnt := inTp.idnt;
    ELSE ASSERT(FALSE);
    END;
    D.AppendIdnt(this.work, idnt);
  END AddNewRecEmitter;

(* ============================================================ *)

  PROCEDURE^ (e : LightningEmitter)EmitProc(proc : Id.Procs; attr : SET),NEW;
  PROCEDURE^ (e : LightningEmitter)PushValue(exp : D.Expr; typ : D.Type),NEW;
  PROCEDURE^ (e : LightningEmitter)PushHandle(exp : D.Expr; typ : D.Type),NEW;
  PROCEDURE^ (e : LightningEmitter)PushRef(exp : D.Expr; typ : D.Type),NEW;
  PROCEDURE^ (e : LightningEmitter)
				EmitStat(stat : D.Stmt; OUT ok : BOOLEAN),NEW;
  PROCEDURE^ (e : LightningEmitter)PushCall(callX : Xp.CallX),NEW;
  PROCEDURE^ (e : LightningEmitter)FallFalse(exp : D.Expr; tLb : INTEGER),NEW;

  PROCEDURE^ (e : LightningEmitter)RefRecCopy(typ : Ty.Record),NEW;
  PROCEDURE^ (e : LightningEmitter)RefArrCopy(typ : Ty.Array),NEW;

(* ============================================================ *)

  PROCEDURE (t : LightningEmitter)NoArgInit(rec : Ty.Record),NEW;
    VAR out : Vos.IlasmFile;
	idx : INTEGER;
	fld : D.Idnt;
  BEGIN
    out := t.outF;
    out.proc := Vos.newProcInfo(t.mod);
    out.Blank();
    out.InitHead(rec);
   (*
    *    Initialize fields, as necessary.
    *)
    IF rec # NIL THEN
      FOR idx := 0 TO rec.fields.tide-1 DO
        fld := rec.fields.a[idx];
	IF ~Vos.hasValueRep(fld.type) THEN
          out.CommentT("Initialize embedded object");
	  out.Code(Asm.opc_ldarg_0);
	  out.StructInit(fld);
        END;
      END;
    END;
    out.Code(Asm.opc_ret);
    out.InitTail(rec);
  END NoArgInit;

(* ============================================================ *)

  PROCEDURE (t : LightningEmitter)CopyProc(recT : Ty.Record),NEW;
    VAR out  : Vos.IlasmFile;
	indx : INTEGER;
	fTyp : D.Type;
	idnt : Id.FldId;
  BEGIN
   (*
    *   Emit the copy procedure "__copy__()
    *)
    out := t.outF;
    out.Blank();
    out.CopyHead(recT);
   (*
    *    Recurse to super class, if necessary.
    *) 
    IF (recT.baseTp # NIL) & (recT.baseTp # G.anyRec) THEN
      out.Code(Asm.opc_ldarg_0);
      out.Code(Asm.opc_ldarg_1);
      t.RefRecCopy(recT.baseTp(Ty.Record));
    END;
   (*
    *    Emit field-by-field copy.
    *)
    FOR indx := 0 TO recT.fields.tide-1 DO
      idnt := recT.fields.a[indx](Id.FldId);
      fTyp := idnt.type;
      out.Code(Asm.opc_ldarg_0);
      IF Vos.hasValueRep(fTyp) THEN
	out.Code(Asm.opc_ldarg_1);
	out.PutGetF(Asm.opc_ldfld, idnt);
        out.PutGetF(Asm.opc_stfld, idnt);
      ELSE
	out.PutGetF(Asm.opc_ldfld, idnt);
	out.Code(Asm.opc_ldarg_1);
	out.PutGetF(Asm.opc_ldfld, idnt);
        WITH fTyp : Ty.Array DO 
	    t.RefArrCopy(fTyp);
        | fTyp : Ty.Record DO
	    t.RefRecCopy(fTyp);
	END;
      END;
    END;
    out.Code(Asm.opc_ret);
    out.CopyTail;
  END CopyProc;

(* ============================================================ *)

  PROCEDURE (this : LightningEmitter)
			EmitRecBody(out : Vos.IlasmFile; typId : Id.TypId),NEW;
  (** Create the assembler for a class file for this record. *)
    VAR index  : INTEGER;
	ident  : D.Idnt;
	baseT  : D.Type;
	field  : Id.FldId;
	method : Id.MthId;
	attSet : SET;
	clsSet : SET;
	mthSet : SET;
	record : Ty.Record;
	impRec : Id.TypId;
	valRec : BOOLEAN;
	noCopy : BOOLEAN;
  BEGIN
    out.Blank();
    record := typId.type.boundRecTp()(Ty.Record);
    valRec := ~(D.clsTp IN record.xAttr);
    noCopy := FALSE; 	(* TEMPORARY PENDING OPTIMIZATION *)
   (*
    *   Account for the record attributes.
    *)
    CASE record.recAtt OF
    | Ty.noAtt : attSet := Asm.att_sealed;
    | Ty.isAbs : attSet := Asm.att_abstract;
    | Ty.limit : attSet := Asm.att_empty;
    | Ty.extns : attSet := Asm.att_empty;
    END;
   (*
    *   Account for the identifier visibility.
    *   It appears that the VOS only supports two kinds:
    *		"public" == exported from this assembly
    *		<empty>  == not exported from this assembly
    *	Note that private is enforced by the name mangling
    *   for types that are local to a procedure.
    *)
    IF typId.vMod = D.pubMode THEN
      attSet := attSet + Asm.att_public;
    END;
    clsSet := attSet;
    IF valRec THEN attSet := attSet + Asm.att_value END;
    out.Comment("RECORD " + record.name()^);
    out.clsN := record.xName;
    out.DirectiveIS(Asm.dot_class, attSet, out.clsN);
   (*
    *   Compute the super class attribute.
    *)
    baseT := record.baseTp;
    WITH baseT : Ty.Record DO
      IF baseT.xName = NIL THEN Vos.MkRecName(baseT) END;
      out.DirectiveS(Asm.dot_super, baseT.scopeNm);
    ELSE
      IF valRec THEN
        out.DirectiveS(Asm.dot_super, Vos.valueType);
      ELSE (* use the empty default *)
      END;
    END;
    IF record.interfaces.tide > 0 THEN
      FOR index := 0 TO record.interfaces.tide-1 DO
	impRec := record.interfaces.a[index](Id.TypId);
	baseT  := impRec.type.boundRecTp();
	out.DirectiveS(Asm.dot_implements, baseT(Ty.Record).scopeNm);
      END;
    END;
    out.OpenBrace(2);
   (*
    *  Emit all the fields ...
    *)
    FOR index := 0 TO record.fields.tide-1 DO
      ident := record.fields.a[index];
      field := ident(Id.FldId);
      out.EmitField(field, Asm.att_empty);
    END;
    IF ~valRec THEN this.NoArgInit(record) END;
    IF ~noCopy THEN this.CopyProc(record) END;
   (*
    *  Emit all the (non-forward) methods ...
    *)
    FOR index := 0 TO record.methods.tide-1 DO
      ident  := record.methods.a[index];
      method := ident(Id.MthId);
      IF method.kind = Id.conMth THEN
	mthSet := method.mthAtt * Id.mask;
       (*
	*  Get the extension bits:
	*     {} == att_final ==> inextensible, ie. final AND virtual
	*    {1} == att_isAbs ==> abstract AND virtual
	*    {2} == att_empty ==> empty, and thus virtual
	*  {1,2} == att_extns ==> extensible, thus virtual
	*)
        IF mthSet = {} THEN 
	  IF Id.newBit IN method.mthAtt THEN
	    attSet :=  Asm.att_instance;
	  ELSE
	    attSet :=  Asm.att_final + Asm.att_virtual;
	  END;
        ELSIF mthSet = Id.isAbs THEN 
	  attSet := Asm.att_virtual + Asm.att_abstract;
	ELSE
	  attSet := Asm.att_virtual;
        END;
        this.EmitProc(method, attSet)
      END;
    END;
    out.CloseBrace(2);
    IF valRec THEN out.EmitBoxClass(record, clsSet, this.mod) END;
  END EmitRecBody;

(* ============================================================ *)

  PROCEDURE (this : LightningEmitter)EmitModBody
			(out : Vos.IlasmFile; mod : Id.BlkId),NEW;
  (** Create the assembler for a class file for this module. *)
    VAR index : INTEGER;
	proc  : Id.Procs;
	recT  : D.Type;
	varId : D.Idnt;
	returned : BOOLEAN;
  BEGIN
    out.clsN := mod.xName;
    out.DirectiveIS(Asm.dot_class, modAttrib, out.clsN);
    out.OpenBrace(2);
    FOR index := 0 TO this.mod.procs.tide-1 DO
     (*
      *  Create the mangled name for all procedures
      *)
      proc := this.mod.procs.a[index];
      Vos.MkProcName(proc);
      Vos.RenumberLocals(proc);
    END;
   (* 
    *  Emit all of the static fields
    *)
    FOR index := 0 TO this.mod.locals.tide-1 DO
      varId := this.mod.locals.a[index];
      out.EmitField(varId, Asm.att_static);
    END;
   (*
    *  No constructor for the module "class",
    *  there are never any instances created.
    *)
    out.proc := Vos.newProcInfo(this.mod);
    asmExe := this.mod.main;		(* Bool for assembler *)
    IF this.mod.main THEN
     (*
      *   Emit <clinit>, and module body as main() 
      *)
      out.Blank();
      out.ClinitHead();
      out.InitVars(this.mod);
      out.Code(Asm.opc_ret);
      out.ClinitTail();
      out.Blank();
      out.MainHead();
      this.EmitStat(this.mod.modBody, returned);
      IF returned THEN 
	out.Comment("Continuing directly to CLOSE");
	this.EmitStat(this.mod.modClose, returned);
        IF returned THEN out.Code(Asm.opc_ret) END;
      END;
      out.MainTail();
    ELSE
     (*
      *   Emit single <clinit> incorporating module body
      *)
      out.ClinitHead();
      out.InitVars(this.mod);
      this.EmitStat(this.mod.modBody, returned);
      IF returned THEN out.Code(Asm.opc_ret) END;
      out.ClinitTail();
    END;
   (* 
    *  Emit all of the static procedures
    *)
    out.Blank();
    FOR index := 0 TO this.mod.procs.tide-1 DO
      proc := this.mod.procs.a[index];
      IF (proc.kind = Id.conPrc) &
	 (proc.dfScp.kind = Id.modId) THEN this.EmitProc(proc, staticAtt) END;
    END;
   (* 
    *  And now, just in case exported types have been missed ...
    *)
    FOR index := 0 TO this.mod.expRecs.tide-1 DO
      recT := this.mod.expRecs.a[index];
      IF recT(Ty.Record).xName = NIL THEN Vos.MkRecName(recT(Ty.Record)) END;
    END;
    out.CloseBrace(2);
  END EmitModBody;

(* ============================================================ *)
(*  Mainline emitter, consumes worklist emitting assembler	*)
(*  files until the worklist is empty.				*)
(* ============================================================ *)

  PROCEDURE (this : LightningEmitter)Emit*();
  (** Create the assembler for a class file for this 
      module: overrides EMPTY method in ClassMaker 	*)
    VAR fileName  : FileNames.NameString;
	out       : Vos.IlasmFile;
	classIx   : INTEGER;
	idDesc    : D.Idnt;
	impElem   : Id.BlkId;
  BEGIN
    Vos.MkBlkName(this.mod);
   (*
    *  Create the asmFile structure, and open the output file.
    *)
    out := Vos.newIlasmFile();
    this.outF := out;
    fileName  := this.mod.xName^ + ".il";
    out.file  := GPTextFiles.createFile(fileName);
    IF out.file = NIL THEN
      CPascalS.SemError.Report(177, 0, 0);
      Error.WriteString("Cannot create out-file <" + fileName + ">");
      Error.WriteLn;
      RETURN;
    ELSE
      IF CompState.verbose THEN CompState.Message("Created "+fileName) END;
      out.Header(CompState.srcNam);
      IF this.mod.main THEN out.Comment("This module implements CPmain") END;
      out.Blank();
      out.DirectiveISS(Asm.dot_assembly, {}, 
				"'" + this.mod.xName^ + "' as ",
				'"' + this.mod.xName^ + '.dll" {}');
     (*
      *  Build 1812.10 required the next line.
      *  In the Beta 2204.21 mscorlib is imported 
      *  automatically, and leaving this line in 
      *  attracts a non-fatal warning message.
      *
      out.DirectiveIS(Asm.dot_assembly, Asm.att_extern, Vos.libVersion);
      *)

      out.DirectiveIS(Asm.dot_assembly, Asm.att_extern, "RTS {}");
      out.ExternList();
      out.Blank();
      out.DirectiveS(Asm.dot_namespace, this.mod.xName);
      out.OpenBrace(0);
      classIx := 0;
      WHILE classIx < this.work.tide DO
	idDesc := this.work.a[classIx];
	WITH idDesc : Id.BlkId DO
	    this.EmitModBody(out, idDesc);
	| idDesc : Id.TypId DO
       	    this.EmitRecBody(out, idDesc);
	END;
	INC(classIx);
      END;
      out.Blank();

      out.CloseBrace(0);
      out.Comment("end namespace " + this.mod.xName^);
      out.Comment("end output produced by gpcp");
     (*
      *   Store the filename for the assembler.
      *)
      asmName := this.mod.xName;
      GPTextFiles.CloseFile(out.file);
    END; 
  END Emit;

(* ============================================================ *)
(*		Shared code-emission methods			*)
(* ============================================================ *)

  PROCEDURE (e : LightningEmitter)EmitProc(proc : Id.Procs; attr : SET),NEW;
    VAR out  : Vos.IlasmFile;
        live : BOOLEAN;
	retn : D.Type;
	indx : INTEGER;
	nest : Id.Procs;
  BEGIN
   (*
    *   Recursively emit nested procedures first.
    *)
    FOR indx := 0 TO proc.nestPs.tide-1 DO 
      nest := proc.nestPs.a[indx];
      IF nest.kind = Id.conPrc THEN e.EmitProc(nest, staticAtt) END;
    END;
    out := e.outF;
    out.proc := Vos.newProcInfo(proc);
    out.Blank();
    out.Comment("PROCEDURE " + proc.prcNm^);
   (*
    *  Compute the method attributes
    *)
    IF proc.vMod = D.pubMode THEN	(* explicitly public *)
      attr := attr + Asm.att_public;
    ELSIF proc.dfScp IS Id.Procs THEN	(* nested procedure  *)
      attr := attr + Asm.att_private;
    ELSE
      attr := attr + Asm.att_assembly;
    END;
    out.MethodDecl(attr, proc);
    out.OpenBrace(4);
   (*
    *  Output the body if not ABSTRACT
    *)
    IF attr * Asm.att_abstract = {} THEN
     (*
      *  Initialize any locals which need this.
      *)
      out.InitVars(proc);
      IF proc.rescue # NIL THEN out.Try END;
     (*
      *  Finally! Emit the method body.
      *)
      e.EmitStat(proc.body, live);
     (*
      *  For proper procedure which reach the fall-
      *  through ending just return.
      *)
      IF live & proc.type.isProperProcType() THEN
	out.DoReturn;
      END;
      IF proc.rescue # NIL THEN
        out.Catch(proc);
        e.EmitStat(proc.rescue, live);
        IF live & proc.type.isProperProcType() THEN
	  out.DoReturn;
        END;
        out.EndCatch;
      END;
    ELSE
      out.Comment("Abstract method");
    END;
    out.MethodTail(proc);
  END EmitProc;

(* ============================================================ *)
(*		    Expression Handling Methods			*)
(* ============================================================ *)

  PROCEDURE longValue(lit : D.Expr) : LONGINT;
  BEGIN
    RETURN lit(Xp.LeafX).value.long();
  END longValue;

  PROCEDURE intValue(lit : D.Expr) : INTEGER;
  BEGIN
    RETURN lit(Xp.LeafX).value.int();
  END intValue;

(* ============================================================ *)

  PROCEDURE (e : LightningEmitter)DoCmp(cmpE : INTEGER;
				        tLab : INTEGER;
				        type : D.Type),NEW;
   (**  Compare two TOS elems and jump to tLab if true. *)
   (* ------------------------------------------------- *)
    VAR out  : Vos.IlasmFile;
   (* ------------------------------------------------- *)
    PROCEDURE test(t : INTEGER) : INTEGER;
    BEGIN
      CASE t OF
      | Xp.greT   : RETURN Asm.opc_bgt;
      | Xp.greEq  : RETURN Asm.opc_bge;
      | Xp.notEq  : RETURN Asm.opc_bne_un;
      | Xp.lessEq : RETURN Asm.opc_ble;
      | Xp.lessT  : RETURN Asm.opc_blt;
      | Xp.equal  : RETURN Asm.opc_beq;
      END;
    END test;
   (* ------------------------------------------------- *)
  BEGIN
    out  := e.outF;
    IF (type IS Ty.Base) &
       ((type = G.strTp) OR (type = G.sStrTp)) OR
       ~(type IS Ty.Base) & type.isCharArrayType() THEN
     (*
      *  For strings and character arrays, we simply
      *  call the compare function, then compare the
      *  result with zero. Instructions are polymorphic.
      *)
      out.StaticCall(Vos.aaStrCmp, -1);
      IF cmpE = Xp.equal THEN
        out.CodeLb(Asm.opc_brfalse, tLab);
      ELSIF cmpE = Xp.notEq THEN
        out.CodeLb(Asm.opc_brtrue, tLab);
      ELSE
        out.PushInt(0);
        out.CodeLb(test(cmpE), tLab);
      END;
    ELSE
      out.CodeLb(test(cmpE), tLab);
    END;
  END DoCmp;

(* ---------------------------------------------------- *)

  PROCEDURE (e : LightningEmitter)BinCmp(exp : D.Expr; 
				         tst : INTEGER;
				         lab : INTEGER),NEW;
    VAR binOp : Xp.BinaryX;
  BEGIN
    binOp := exp(Xp.BinaryX);
    e.PushValue(binOp.lKid, binOp.lKid.type);
    IF binOp.lKid.type = G.strTp THEN
     (*
      *  If this is a string, convert to a character array.
      *)
      e.outF.StaticCall(Vos.vStr2ChO, 0);
    END;

    e.PushValue(binOp.rKid, binOp.rKid.type);
    IF binOp.rKid.type = G.strTp THEN
     (*
      *  If this is a string, convert to a character array.
      *)
      e.outF.StaticCall(Vos.vStr2ChO, 0);
    END;
    e.DoCmp(tst, lab, binOp.lKid.type);
  END BinCmp;

(* ---------------------------------------------------- *)

  PROCEDURE (e : LightningEmitter)FallTrue(exp : D.Expr; fLb : INTEGER),NEW;
   (** Evaluate exp, fall through if true, jump to fLab otherwise *)
    VAR binOp : Xp.BinaryX;
	label : INTEGER;
	out   : Vos.IlasmFile;
  BEGIN
    out := e.outF;
    CASE exp.kind OF
    | Xp.tBool :				(* just do nothing *)
    | Xp.fBool : 
	out.CodeLb(Asm.opc_br, fLb);
    | Xp.blNot :
	e.FallFalse(exp(Xp.UnaryX).kid, fLb);
    | Xp.greT :
	e.BinCmp(exp, Xp.lessEq, fLb);
    | Xp.greEq :
	e.BinCmp(exp, Xp.lessT, fLb);
    | Xp.notEq :
	e.BinCmp(exp, Xp.equal, fLb);
    | Xp.lessEq :
	e.BinCmp(exp, Xp.greT, fLb);
    | Xp.lessT :
	e.BinCmp(exp, Xp.greEq, fLb);
    | Xp.equal :
	e.BinCmp(exp, Xp.notEq, fLb);
    | Xp.blOr :
	binOp := exp(Xp.BinaryX);
	label := out.newLabel();
	e.FallFalse(binOp.lKid, label);
	e.FallTrue(binOp.rKid, fLb);
	out.DefLab(label);
    | Xp.blAnd :
	binOp := exp(Xp.BinaryX);
	e.FallTrue(binOp.lKid, fLb);
	e.FallTrue(binOp.rKid, fLb);
    | Xp.isOp :
	binOp := exp(Xp.BinaryX);
	e.PushValue(binOp.lKid, binOp.lKid.type);
	out.CodeT(Asm.opc_isinst, binOp.rKid(Xp.IdLeaf).ident.type);
	(* if NIL then FALSE... *)
	out.CodeLb(Asm.opc_brfalse, fLb);
    | Xp.inOp :
	binOp := exp(Xp.BinaryX);
	out.Code(Asm.opc_ldc_i4_1);
	e.PushValue(binOp.lKid, binOp.lKid.type);
	out.Code(Asm.opc_shl);
	out.Code(Asm.opc_dup);
	e.PushValue(binOp.rKid, binOp.rKid.type);
	out.Code(Asm.opc_and);
	out.CodeLb(Asm.opc_bne_un, fLb);
    ELSE (* Xp.fnCll, Xp.qualId, Xp.index, Xp.selct  *)
      e.PushValue(exp, exp.type);		(* boolean variable *)
      out.CodeLb(Asm.opc_brfalse, fLb);
    END;
  END FallTrue;

(* ---------------------------------------------------- *)

  PROCEDURE (e : LightningEmitter)FallFalse(exp : D.Expr; tLb : INTEGER),NEW;
   (** Evaluate exp, fall through if false, jump to tLb otherwise *)
    VAR binOp : Xp.BinaryX;
	label : INTEGER;
	out   : Vos.IlasmFile;
  BEGIN
    out := e.outF;
    CASE exp.kind OF
    | Xp.fBool :				(* just do nothing *)
    | Xp.tBool : 
	out.CodeLb(Asm.opc_br, tLb);
    | Xp.blNot :
	e.FallTrue(exp(Xp.UnaryX).kid, tLb);
    | Xp.greT :
	e.BinCmp(exp, Xp.greT, tLb);
    | Xp.greEq :
	e.BinCmp(exp, Xp.greEq, tLb);
    | Xp.notEq :
	e.BinCmp(exp, Xp.notEq, tLb);
    | Xp.lessEq :
	e.BinCmp(exp, Xp.lessEq, tLb);
    | Xp.lessT :
	e.BinCmp(exp, Xp.lessT, tLb);
    | Xp.equal :
	e.BinCmp(exp, Xp.equal, tLb);
    | Xp.blOr :
	binOp := exp(Xp.BinaryX);
	e.FallFalse(binOp.lKid, tLb);
	e.FallFalse(binOp.rKid, tLb);
    | Xp.blAnd :
	label := out.newLabel();
	binOp := exp(Xp.BinaryX);
	e.FallTrue(binOp.lKid, label);
	e.FallFalse(binOp.rKid, tLb);
	out.DefLab(label);
    | Xp.isOp :
	binOp := exp(Xp.BinaryX);
	e.PushValue(binOp.lKid, binOp.lKid.type);
	out.CodeT(Asm.opc_isinst, binOp.rKid(Xp.IdLeaf).ident.type);
	(* if non-NIL then TRUE... *)
	out.CodeLb(Asm.opc_brtrue, tLb);
    | Xp.inOp :
	binOp := exp(Xp.BinaryX);
	out.Code(Asm.opc_ldc_i4_1);
	e.PushValue(binOp.lKid, binOp.lKid.type);
	out.Code(Asm.opc_shl);
	out.Code(Asm.opc_dup);
	e.PushValue(binOp.rKid, binOp.rKid.type);
	out.Code(Asm.opc_and);
	out.CodeLb(Asm.opc_beq, tLb);
    ELSE (* Xp.fnCll, Xp.qualId, Xp.index, Xp.selct  *)
      e.PushValue(exp, exp.type);		(* boolean variable *)
      out.CodeLb(Asm.opc_brtrue, tLb);
    END;
  END FallFalse;

(* ============================================================ *)

  PROCEDURE (e : LightningEmitter)PushUnary(exp : Xp.UnaryX; dst : D.Type),NEW;
    VAR dNum : INTEGER;
	code : INTEGER;
	labl : INTEGER;
	out  : Vos.IlasmFile;
	ovfl : BOOLEAN;
  BEGIN
    e.PushValue(exp.kid, exp.kid.type);
    out := e.outF;
    ovfl := out.proc.prId.ovfChk;
    CASE exp.kind OF
    | Xp.mkStr  : (* skip *)
    | Xp.deref  : 
	IF Vos.isValRecord(dst) THEN out.GetVal(exp.kid.type(Ty.Pointer)) END;
    | Xp.mkNStr :
	IF exp.kid.type # G.strTp THEN out.StaticCall(Vos.chs2Str, 0) END;
    | Xp.strChk :
	out.Code(Asm.opc_dup);
	out.StaticCall(Vos.aStrChk, -1);	(* Do some range checks *)
    | Xp.compl :
	out.Code(Asm.opc_ldc_i4_M1);
	out.Code(Asm.opc_xor);
    | Xp.neg :
	out.Code(Asm.opc_neg);
    | Xp.absVl :
	dNum := dst(Ty.Base).tpOrd;
	IF ~ovfl THEN
	 (*
	  *  This is the code to use for non-trapping cases
	  *)
	  out.Code(Asm.opc_dup);
	  out.Code(Asm.opc_ldc_i4_0);
	  IF    dNum = Ty.realN THEN
	    out.Code(Asm.opc_conv_r8);
	  ELSIF dNum = Ty.sReaN THEN
	    out.Code(Asm.opc_conv_r4);
	  ELSIF dNum = Ty.lIntN THEN
	    out.Code(Asm.opc_conv_i8);
	  (* ELSE do nothing for all INTEGER cases *)
	  END;
	  labl := out.newLabel();
	  out.CodeLb(Asm.opc_ble, labl);
	  out.Code(Asm.opc_neg);
	  out.DefLab(labl);
	ELSE
	 (*
	  *  The following is the safe but slow code.
	  *)
	  IF    dNum = Ty.realN THEN
	    out.StaticCall(Vos.dAbs, 0);	
	  ELSIF dNum = Ty.sReaN THEN
	    out.StaticCall(Vos.fAbs, 0);	
	  ELSIF dNum = Ty.lIntN THEN
	    out.StaticCall(Vos.lAbs, 0);	
          ELSE
	    out.StaticCall(Vos.iAbs, 0);	
          END;
        END;
    | Xp.entVl :
	dNum := dst(Ty.Base).tpOrd;
	IF dNum = Ty.sReaN THEN out.Code(Asm.opc_conv_r8) END;
	(*
	// We _could_ check if the value is >= 0.0, and 
	// skip the call in that case, falling through
	// into the round-to-zero mode opc_d2l.
	*)
	out.StaticCall(Vos.dFloor, 0);	
	IF ~ovfl THEN
	  out.Code(Asm.opc_conv_i8);
	ELSE
	  out.Code(Asm.opc_conv_ovf_i8);
	END;
    | Xp.capCh :
	out.StaticCall(Vos.toUpper, 0);	
    | Xp.blNot :
	out.Code(Asm.opc_ldc_i4_1);
	out.Code(Asm.opc_xor); 
    | Xp.strLen :
	out.StaticCall(Vos.aStrLen, 0);	
    | Xp.oddTst :
	out.Code(Asm.opc_ldc_i4_1);
	out.Code(Asm.opc_and); 
    END;
  END PushUnary;

(* ============================================================ *)

  PROCEDURE (e : LightningEmitter)
			PushBinary(exp : Xp.BinaryX; dst : D.Type),NEW;
    VAR out  : Vos.IlasmFile;
	lOp  : D.Expr;
	rOp  : D.Expr;
	dNum : INTEGER;
	sNum : INTEGER;
	code : INTEGER;
	indx : INTEGER;
	rLit : LONGINT;
	exLb : INTEGER;
	tpLb : INTEGER;
	long : BOOLEAN;
	temp : INTEGER;
	ovfl : BOOLEAN;
  BEGIN
    out := e.outF;
    lOp := exp.lKid;
    rOp := exp.rKid;
    ovfl := out.proc.prId.ovfChk & dst.isIntType();
    CASE exp.kind OF
    (* -------------------------------- *)
    | Xp.index :
	e.PushValue(lOp, lOp.type);		(* push array designator*)
	e.PushValue(rOp, rOp.type);		(* push index value	*)
        IF Vos.isValRecord(dst) THEN out.CodeTn(Asm.opc_ldelema, dst) END;
	out.GetElem(dst);			(* load the element	*)
    (* -------------------------------- *)
    | Xp.range : 				(* set i..j range ...	*)
       (* We want to create an integer with bits--	*)
       (*      [0...01...10...0]			*)
       (* MSB==31    j   i    0==LSB			*)
       (* One method is A				*)
       (* 1)   [0..010........0]  1 << (j+1)		*)
       (* 2)   [1..110........0]  negate(1)		*)
       (* 3)   [0.......010...0]  1 << i		*)
       (* 4)   [1.......110...0]  negate(3)		*)
       (* 5)   [0...01...10...0]  (2)xor(4)		*)
       (* Another method is B				*)
       (* 1)   [1.............1]  -1			*)
       (* 2)   [0...01........1]  (1) >>> (31-j)	*)
       (* 3)   [0........01...1]  (2) >> i		*)
       (* 4)   [0...01...10...0]  (3) << i		*)
       (* --------------------------------------------- *
	*      (*					*
	*	* Method A				*
	*	*)					*
	*	out.Code(Asm.opc_ldc_i4_1);		*
	*	out.Code(Asm.opc_ldc_i4_1);		*
	*	e.PushValue(rOp, G.intTp);		*
	*      (* Do unsigned less than 32 test here *)	*
	*	out.Code(Asm.opc_add);			*
	*	out.Code(Asm.opc_shl);			*
	*	out.Code(Asm.opc_neg);			*
	*	out.Code(Asm.opc_ldc_i4_1);		*
	*	e.PushValue(lOp, G.intTp);		*
	*      (* Do unsigned less than 32 test here *) *
	*	out.Code(Asm.opc_shl);			*
	*	out.Code(Asm.opc_neg);			*
	*	out.Code(Asm.opc_xor);			*
	* -------------------------------------------- *)
       (*
	* Method B
	*)
	IF rOp.kind = Xp.numLt THEN
	  indx := intValue(rOp);
	  IF indx = 31 THEN
	    out.Code(Asm.opc_ldc_i4_M1);
	  ELSE
	    temp := ORD({0 .. indx});
	    out.PushInt(temp);
	  END;
	ELSE
	  out.Code(Asm.opc_ldc_i4_M1);
	  out.PushInt(31);
	  e.PushValue(rOp, G.intTp);
	 (* Do unsigned less than 32 test here ...*)
	  out.Code(Asm.opc_sub);
	  out.Code(Asm.opc_shr_un);
	END;
	IF lOp.kind = Xp.numLt THEN
	  indx := intValue(lOp);
	  IF indx > 0 THEN
	    temp := ORD({indx .. 31});
	    out.PushInt(temp);
	    out.Code(Asm.opc_and);
	  END;
	ELSE
	  e.PushValue(lOp, G.intTp);
	 (* Do unsigned less than 32 test here ...*)
	  out.Code(Asm.opc_dup);
	  temp := out.proc.newLocal(G.intTp);
	  out.StoreLocal(temp);
	  out.Code(Asm.opc_shr);
	  out.LoadLocal(temp);
	  out.Code(Asm.opc_shl);
          out.proc.ReleaseLocal(temp);
	END;
    (* -------------------------------- *)
    | Xp.lenOf :
	e.PushValue(lOp, lOp.type);
	FOR indx := 0 TO intValue(rOp) - 1 DO
	  out.Code(Asm.opc_ldc_i4_0);
	  out.Code(Asm.opc_ldelem_ref);
	END;
	out.Code(Asm.opc_ldlen);
    (* -------------------------------- *)
    | Xp.maxOf, Xp.minOf :
	tpLb := out.newLabel();
	exLb := out.newLabel();
       (*
	* Push left operand, duplicate
	* stack is (top) lOp lOp...
	*)
	e.PushValue(lOp, dst);
	out.Code(Asm.opc_dup);
       (*
        * Push right operand, duplicate
        * stack is (top) rOp rOp lOp lOp ...
	*)
	e.PushValue(rOp, dst); 
	out.Code(Asm.opc_dup);
       (*
        * Store rOp to temp
        * stack is (top) rOp lOp lOp ...
	*)
	temp := out.proc.newLocal(dst);
        out.StoreLocal(temp);
       (*
	*  Compare two top items and jump
	*  stack is (top) lOp ...
	*)
	IF exp.kind = Xp.maxOf THEN
	  e.DoCmp(Xp.greT, exLb, dst);		(* leaving lOp on stack *)
	ELSE
	  e.DoCmp(Xp.lessT, exLb, dst);		(* leaving lOp on stack *)
	END;
       (*
	*  Else: discard top item
	*  and push stored rOp instead
	*)
	out.Code(Asm.opc_pop);
        out.LoadLocal(temp);
	out.DefLab(exLb);
        out.proc.ReleaseLocal(temp);
    (* -------------------------------- *)
    | Xp.bitAnd :
	e.PushValue(lOp, dst);
	e.PushValue(rOp, dst);
	out.Code(Asm.opc_and);
    (* -------------------------------- *)
    | Xp.bitOr :
	e.PushValue(lOp, dst);
	e.PushValue(rOp, dst);
	out.Code(Asm.opc_or);
    (* -------------------------------- *)
    | Xp.bitXor :
	e.PushValue(lOp, dst);
	e.PushValue(rOp, dst);
	out.Code(Asm.opc_xor);
    (* -------------------------------- *)
    | Xp.plus :
	e.PushValue(lOp, dst);
	e.PushValue(rOp, dst);
	IF ovfl THEN out.Code(Asm.opc_add_ovf) ELSE out.Code(Asm.opc_add) END;
    (* -------------------------------- *)
    | Xp.minus :
	e.PushValue(lOp, dst);
	e.PushValue(rOp, dst);
	IF ovfl THEN out.Code(Asm.opc_sub_ovf) ELSE out.Code(Asm.opc_sub) END;
    (* -------------------------------- *)
    | Xp.mult :
	e.PushValue(lOp, dst);
	e.PushValue(rOp, dst);
	IF ovfl THEN out.Code(Asm.opc_mul_ovf) ELSE out.Code(Asm.opc_mul) END;
    (* -------------------------------- *)
    | Xp.slash :
	e.PushValue(lOp, dst);
	e.PushValue(rOp, dst);
	out.Code(Asm.opc_div);
    (* -------------------------------- *)
    | Xp.rem0op :
	dNum := dst(Ty.Base).tpOrd;
	e.PushValue(lOp, dst);
	e.PushValue(rOp, dst);
	out.Code(Asm.opc_rem);
    (* -------------------------------- *)
    | Xp.modOp :
	long := dst(Ty.Base).tpOrd = Ty.lIntN;
	e.PushValue(lOp, dst);
	e.PushValue(rOp, dst);
	IF (rOp.kind = Xp.numLt) & (intValue(rOp) > 0) THEN
	  indx := intValue(rOp);
	  tpLb := out.newLabel();
	  out.Code(Asm.opc_rem);
	  out.Code(Asm.opc_dup);
	  IF long THEN out.PushLong(0) ELSE out.PushInt(0) END;
	  out.CodeLb(Asm.opc_bge, tpLb);
	  IF long THEN out.PushLong(indx) ELSE out.PushInt(indx) END;
	  out.Code(Asm.opc_add);
	  out.DefLab(tpLb);
	ELSIF long THEN
	  out.StaticCall(Vos.CpModL, -1);
	ELSE
	  out.StaticCall(Vos.CpModI, -1);
	END;
    (* -------------------------------- *)
    | Xp.div0op :
	dNum := dst(Ty.Base).tpOrd;
	e.PushValue(lOp, dst);
	e.PushValue(rOp, dst);
	out.Code(Asm.opc_div);
    (* -------------------------------- *)
    | Xp.divOp :
	long := dst(Ty.Base).tpOrd = Ty.lIntN;
	e.PushValue(lOp, dst);
	IF (rOp.kind = Xp.numLt) & (longValue(rOp) > 0) THEN
	  tpLb := out.newLabel();
	  out.Code(Asm.opc_dup);
	  IF long THEN
	    rLit := longValue(rOp);
	    out.PushLong(0);
	    out.CodeLb(Asm.opc_bge, tpLb);
	    out.PushLong(rLit-1);
	    out.Code(Asm.opc_sub);
	    out.DefLab(tpLb);
	    out.PushLong(rLit);
	  ELSE
	    indx := intValue(rOp);
	    out.PushInt(0);
	    out.CodeLb(Asm.opc_bge, tpLb);
	    out.PushInt(indx-1);
	    out.Code(Asm.opc_sub);
	    out.DefLab(tpLb);
	    out.PushInt(indx);
	  END;
	  out.Code(Asm.opc_div);
	ELSE
	  e.PushValue(rOp, dst);
	  IF long THEN
	    out.StaticCall(Vos.CpDivL, -1);
	  ELSE
	    out.StaticCall(Vos.CpDivI, -1);
	  END;
	END;
    (* -------------------------------- *)
    | Xp.blOr,  Xp.blAnd,  Xp.greT,  Xp.greEq, 
      Xp.notEq, Xp.lessEq, Xp.lessT, Xp.equal, Xp.inOp :
	tpLb := out.newLabel();
	exLb := out.newLabel();
       (* 
	*  Jumping code is mandated for blOr and blAnd...
	* 
	*  For the Relational Ops this next seems crude, but
	*  appears to be the only way that the JVM allows
	*  construction of boolean values.
	*)
	e.FallTrue(exp, tpLb);
	out.Code(Asm.opc_ldc_i4_1);
	out.CodeLb(Asm.opc_br, exLb);
	out.DefLab(tpLb);
	out.Code(Asm.opc_ldc_i4_0);
	out.DefLab(exLb);
    (* -------------------------------- *)
    | Xp.isOp :
	e.PushValue(lOp, lOp.type);
	out.CodeT(Asm.opc_isinst, rOp(Xp.IdLeaf).ident.type);
	out.Code(Asm.opc_ldnull);
	out.Code(Asm.opc_cgt_un);
    (* -------------------------------- *)
    | Xp.ashInt :
	e.PushValue(lOp, lOp.type);
	IF rOp.kind = Xp.numLt THEN
	  indx := intValue(rOp);
	  IF indx = 0 THEN (* skip *)
	  ELSIF indx < 0 THEN
	    out.PushInt(-indx);
	    out.Code(Asm.opc_shr);
	  ELSE
	    out.PushInt(indx);
	    out.Code(Asm.opc_shl);
	  END;
	ELSE
	  tpLb := out.newLabel();
	  exLb := out.newLabel();
	 (*
	  *  This is a variable shift. Do it the hard way.
	  *  First, check the sign of the right hand op.
	  *)
	  e.PushValue(rOp, rOp.type);
	  out.Code(Asm.opc_dup);
	  out.PushInt(0);
	  out.CodeLb(Asm.opc_blt, tpLb);
	 (*
	  *  Positive selector ==> shift left;
	  *)
	  out.Code(Asm.opc_shl);
	  out.CodeLb(Asm.opc_br, exLb);
	 (*
	  *  Negative selector ==> shift right;
	  *)
	  out.DefLab(tpLb);
	  out.Code(Asm.opc_neg);
	  out.Code(Asm.opc_shr);
	  out.DefLab(exLb);
	END;
    (* -------------------------------- *)
    | Xp.strCat :
	e.PushValue(lOp, lOp.type);
	e.PushValue(rOp, rOp.type);
	IF lOp.type = G.strTp THEN
	  IF rOp.type = G.strTp THEN
	    out.StaticCall(Vos.CPJstrCatSS, -1);
	  ELSE
	    out.StaticCall(Vos.CPJstrCatSA, -1);
	  END;
	ELSE
	  IF rOp.type = G.strTp THEN
	    out.StaticCall(Vos.CPJstrCatAS, -1);
	  ELSE
	    out.StaticCall(Vos.CPJstrCatAA, -1);
	  END;
	END;
    (* -------------------------------- *)
    END;
  END PushBinary;

(* ============================================================ *)

  PROCEDURE (e : LightningEmitter)PushValue(exp : D.Expr; typ : D.Type),NEW;
    VAR out : Vos.IlasmFile;
	rec : Ty.Record;
	ix  : INTEGER;
	elm : D.Expr;
	emt : BOOLEAN;		(* ==> more than one set element expr *)
  BEGIN
    out := e.outF;
    WITH exp : Xp.IdLeaf DO
	out.GetVar(exp.ident);
    | exp : Xp.SetExp DO
	emt := TRUE;
       (*
	*   Write out the constant part, if there is one.
	*)
	IF exp.value # NIL THEN
	  out.PushInt(exp.value.int()); (* const part *)
	  emt := FALSE;
	END;
       (*
	*   Write out the element expressions.
	*   taking the union with any part emitted already.
	*)
	FOR ix := 0 TO exp.varSeq.tide-1 DO
	  elm := exp.varSeq.a[ix];
	  IF elm.kind = Xp.range THEN
	    e.PushValue(elm, G.intTp);
	  ELSE
	    out.PushInt(1);
	    e.PushValue(exp.varSeq.a[ix], G.intTp);
	    out.Code(Asm.opc_shl);
	  END;
	  IF ~emt THEN out.Code(Asm.opc_or) END;
	  emt := FALSE;
        END;
       (*
	*   If neither of the above emitted anything, emit zero!
	*)
	IF emt THEN out.Code(Asm.opc_ldc_i4_0) END;
    | exp : Xp.LeafX DO
        CASE exp.kind OF
	| Xp.tBool  : out.Code(Asm.opc_ldc_i4_1);
	| Xp.fBool  : out.Code(Asm.opc_ldc_i4_0);
	| Xp.nilLt  : out.Code(Asm.opc_ldnull);
	| Xp.charLt : out.PushInt(ORD(exp.value.char()));
	| Xp.setLt  : out.PushInt(exp.value.int());
	| Xp.numLt  : 
	    IF typ = G.lIntTp THEN
	      out.PushLong(exp.value.long());
	    ELSE
	      out.PushInt(exp.value.int());
	    END;
	| Xp.realLt : 
	    IF typ = G.realTp THEN
	      out.PushReal(exp.value.real());
	    ELSE
	      out.PushSReal(exp.value.real());
	    END;
	| Xp.strLt  :
	    IF typ = G.charTp THEN
	      out.PushInt(ORD(exp.value.chr0()));
	    ELSE
	      out.PushStr(exp.value.chOpen());
	    END;
        END;
    | exp : Xp.CallX DO
  	e.PushCall(exp);
    | exp : Xp.IdentX DO
	IF exp.kind = Xp.selct THEN
	  e.PushHandle(exp, exp.type);
	  rec := exp.kid.type(Ty.Record);
	  out.PutGetF(Asm.opc_ldfld, exp.ident(Id.FldId));
	ELSE
	  e.PushValue(exp.kid, exp.kid.type);
	  IF exp.kind = Xp.cvrtUp THEN
	    out.ConvertUp(exp.kid.type, typ);
	  ELSIF exp.kind = Xp.cvrtDn THEN
	    out.ConvertDn(exp.kid.type, typ);
	  ELSIF exp.kind = Xp.tCheck THEN
	    rec := exp.ident.type.boundRecTp()(Ty.Record);
	    out.CodeT(Asm.opc_castclass, rec);
	  END;
	END;
    | exp : Xp.UnaryX DO
	e.PushUnary(exp, typ);
    | exp : Xp.BinaryX DO
	e.PushBinary(exp, typ);
    END;
  END PushValue;

(* ---------------------------------------------------- *)

  PROCEDURE (e : LightningEmitter)PushHandle(exp : D.Expr; typ : D.Type),NEW;
   (* Pre: 	exp must be a variable designator 		*)
   (* Post:	Reference to containing object is on TOS.	*)
    VAR vble : Id.ParId;
	kidX : D.Expr;
	kidT : D.Type;
	valI : INTEGER;
    VAR out  : Vos.IlasmFile;
  BEGIN
    out := e.outF;
    ASSERT(exp.isVarDesig());
    WITH exp : Xp.IdentX DO
	kidX := exp.kid; 
	kidT := kidX.type;
        IF Vos.isValRecord(kidT) THEN 
	  e.PushRef(kidX,kidT);
	ELSE 
	  e.PushValue(kidX,kidT);
        END;
    | exp : Xp.BinaryX DO
	e.PushValue(exp.lKid, exp.lKid.type);
	e.PushValue(exp.rKid, G.intTp);
        IF Vos.isValRecord(typ) THEN out.CodeTn(Asm.opc_ldelema, typ) END;
    | exp : Xp.IdLeaf DO
	IF ~Vos.hasValueRep(typ) THEN
	  e.PushValue(exp, typ);
	ELSIF exp.ident.kind = Id.parId THEN
         (*
	  *   This is a param of some value type.
	  *   If this is a reference param then
	  *   push the arg, else do nothing.
	  *)
	  vble := exp.ident(Id.ParId);
	  IF vble.boxOrd # D.val THEN out.LoadArg(vble.varOrd) END;
	END;
    | exp : Xp.UnaryX DO
	ASSERT(exp.kind = Xp.deref);
	kidX := exp.kid; 
	kidT := kidX.type;
	e.PushValue(kidX, kidT);
        IF Vos.isValRecord(typ) THEN out.GetValA(exp.kid.type(Ty.Pointer)) END;
    ELSE
      e.PushValue(exp, typ);
    END;
  END PushHandle;

(* ---------------------------------------------------- *)

  PROCEDURE (e : LightningEmitter)PushRef(exp : D.Expr; typ : D.Type),NEW;
    VAR out : Vos.IlasmFile;
	rec : Ty.Record;
  BEGIN
    ASSERT(exp.isVarDesig());
    out := e.outF;
    e.PushHandle(exp, typ);
    WITH exp : Xp.IdLeaf DO	(* A scalar variable *)
	out.GetVarA(exp.ident);
    | exp : Xp.IdentX DO	(* A field reference *)
	out.PutGetF(Asm.opc_ldflda, exp.ident(Id.FldId));
    | exp : Xp.BinaryX DO	(* An array element  *)
        IF ~Vos.isValRecord(typ) THEN e.outF.CodeTn(Asm.opc_ldelema, typ) END;
    ELSE (* skip *)
    END;
  END PushRef;

(* ---------------------------------------------------- *)

  PROCEDURE (e : LightningEmitter)ValueAssign(exp : D.Expr),NEW;
    VAR out : Vos.IlasmFile;
	rec : Ty.Record;
  BEGIN
    out := e.outF;
    WITH exp : Xp.IdLeaf DO
	(* stack has ... value, (top)	*)
	out.PutVar(exp.ident);
    | exp : Xp.IdentX DO
	(* stack has ... obj-ref, value, (top)	*)
	rec := exp.kid.type(Ty.Record);
	out.PutGetF(Asm.opc_stfld, exp.ident(Id.FldId));
    | exp : Xp.BinaryX DO
       (*
	*  IF element type is a value-rec, 
	*  THEN Stack: ... elem-adr, value, (top)
	*  ELSE  ... arr-ref, index, value, (top)
	*)
	out.PutElem(exp.type);
    | exp : Xp.UnaryX DO
       (*
	*  This is a deref of a value record
	*  and Stack: ... handle, value, (top)
	*)
	out.CodeT(Asm.opc_stobj, exp.type);
    ELSE
      Console.WriteString("BAD VALUE ASSIGN"); Console.WriteLn;
      exp.Diagnose(0);
      ASSERT(FALSE);
    END;
  END ValueAssign;

(* ---------------------------------------------------- *)

  PROCEDURE (e : LightningEmitter)RefRecCopy(typ : Ty.Record),NEW;
    VAR nam : LitValue.CharOpen;
  BEGIN
   (*
    *  We should use a builtin here for value type, but
    *  seem to need an element by element copy for classes.
    *
    *     Stack at entry is (top) srcRef, dstRef...
    *)
    IF typ.xName = NIL THEN Vos.MkRecName(typ) END;
    e.outF.CopyCall(typ);
  END RefRecCopy;

(* ---------------------------------------------------- *)

  PROCEDURE (e : LightningEmitter)RefArrCopy(typ : Ty.Array),NEW;
    VAR out   : Vos.IlasmFile;
	count : INTEGER;
	cardN : INTEGER;
	dstLc : INTEGER;
	srcLc : INTEGER;
	sTemp : INTEGER;
	label : INTEGER;
	elTyp : D.Type;
  BEGIN
    elTyp := typ.elemTp;
   (*
    *    Stack at entry is (top) srcRef, dstRef...
    *)
    out := e.outF;
    label := out.newLabel();
   (*
    *   Allocate two local variables.
    *)
    dstLc := out.proc.newLocal(typ);
    srcLc := out.proc.newLocal(typ);
   (*
    *   Initialize the two locals.
    *)
    out.StoreLocal(srcLc);
    out.StoreLocal(dstLc);			(* Stack is now empty.	*)
    cardN := typ.length;
   (*
    *   Compute the length, either now or at runtime...
    *)
    IF  (cardN > 0) & 				(* ... not open array	*)
	(cardN <= inlineLimit) &		(* ... not too long	*)
	 Vos.hasValueRep(elTyp) THEN		(* ... outer dimension  *)
     (*
      *   Do inline <element assign> using a compile-time loop
      *)
      FOR count := 0 TO cardN-1 DO
        out.LoadLocal(dstLc);
        out.PushInt(count);
        IF Vos.isValRecord(elTyp) THEN
          out.CodeTn(Asm.opc_ldelema, elTyp);
          out.LoadLocal(srcLc);
          out.PushInt(count);
          out.CodeTn(Asm.opc_ldelema, elTyp);
          out.CodeT(Asm.opc_ldobj, elTyp);
          out.CodeT(Asm.opc_stobj, elTyp);
        ELSE
          IF (elTyp IS Ty.Array) OR 
	     (elTyp IS Ty.Record) THEN out.GetElem(elTyp) END;
          out.LoadLocal(srcLc);
          out.PushInt(count);
          out.GetElem(elTyp);
          WITH elTyp : Ty.Record DO
	      e.RefRecCopy(elTyp);
          | elTyp : Ty.Array DO
	      e.RefArrCopy(elTyp);
          ELSE  (* scalar element type *)
	    out.PutElem(elTyp);
          END;
        END;
      END;
    ELSE (* Do array copy using a runtime loop *)
      IF cardN = 0 THEN (* open array, get length from source desc *)
        out.LoadLocal(srcLc);
	out.Code(Asm.opc_ldlen);
      ELSE 
	out.PushInt(cardN);
      END;
     (*
      *   Allocate an extra local variable
      *)
      count := out.proc.newLocal(G.intTp);
      out.StoreLocal(count);
      out.DefLab(label);			(* The back-edge target *)
     (*
      *   Decrement the loop count.
      *)
      out.DecTemp(count); 			(* Stack is now empty.	*)
     (*
      *   We now do the one-per-loop <element assign>
      *)
      out.LoadLocal(dstLc);
      out.LoadLocal(count);
      IF Vos.isValRecord(elTyp) THEN
        out.CodeTn(Asm.opc_ldelema, elTyp);
        out.LoadLocal(srcLc);
        out.LoadLocal(count);
        out.CodeTn(Asm.opc_ldelema, elTyp);
        out.CodeT(Asm.opc_ldobj, elTyp);
        out.CodeT(Asm.opc_stobj, elTyp);
      ELSE
        IF (elTyp IS Ty.Array) OR 
	   (elTyp IS Ty.Record) THEN out.GetElem(elTyp) END;
        out.LoadLocal(srcLc);
        out.LoadLocal(count);
        out.GetElem(elTyp);
        WITH elTyp : Ty.Record DO
	    e.RefRecCopy(elTyp);
        | elTyp : Ty.Array DO
	    e.RefArrCopy(elTyp);
        ELSE  (* scalar element type *)
	  out.PutElem(elTyp);
        END;
      END;
     (*
      *   Loop back to label if count non-zero.
      *)
      out.LoadLocal(count);
      out.CodeLb(Asm.opc_brtrue, label);
     (*
      *   release the extra local.
      *)
      out.proc.ReleaseLocal(count);
    END;
   (*
    *   ... and release the two locals.
    *)
    out.proc.ReleaseLocal(srcLc);
    out.proc.ReleaseLocal(dstLc);
  END RefArrCopy;

(* ---------------------------------------------------- *)

  PROCEDURE (e : LightningEmitter)ValueCopy(act : D.Expr; fmT : D.Type),NEW;
   (**  Make a copy of the value of expression act, into a newly 	*)
   (*   allocated destination.  Leave dst reference on top of stack.	*)
    VAR out  : Vos.IlasmFile;
	dTmp : INTEGER;
	sTmp : INTEGER;
  BEGIN
   (*
    *  Copy this actual, where fmT is either an array or record.
    *)
    out := e.outF;
    WITH fmT : Ty.Record DO
      out.MkNewRecord(fmT);			(* (top) dst...		*)
      out.Code(Asm.opc_dup);			(* (top) dst,dst...	*)
      e.PushValue(act, fmT);			(* (top) src,dst,dst...	*)
      e.RefRecCopy(fmT);			(* (top) dst...		*)
    | fmT : Ty.Array DO
     (*
      *  Array case: ordinary value copy
      *)
      dTmp := out.proc.newLocal(fmT);		(* Holds dst reference. *)
      sTmp := out.proc.newLocal(fmT);		(* Holds src reference. *)
      IF fmT.length = 0 THEN 			
       (* 
	*   This is the open array destination case.  
	*   Compute length of actual parameter and
	*   allocate an array of the needed length.
	*)
	e.PushValue(act, fmT);			(* (top) src...		*)
	out.Code(Asm.opc_dup);			(* (top) src,src...	*)
	out.StoreLocal(sTmp);			(* (top) src...		*)

	IF act.kind = Xp.mkStr THEN		(* (top) src...		*)
	  out.StaticCall(Vos.aStrLp1, 0);	(* (top) len...		*)
	  out.CodeTn(Asm.opc_newarr, G.charTp);	(* (top) dst...		*)
	ELSE					(* (top) src...		*)
	  out.MkArrayCopy(fmT);			(* (top) dst...		*)
	END;
	out.Code(Asm.opc_dup);			(* (top) dst,dst...	*)
	out.StoreLocal(dTmp);			(* (top) dst...		*)
	out.LoadLocal(sTmp);			(* (top) src,dst...	*)
      ELSE 
       (* 
	*   This is the fixed-length destination case.
	*   We allocate an array of the needed length.
	*)
	out.MkFixedArray(fmT);	
	out.Code(Asm.opc_dup);			(* (top) dst,dst...	*)
	out.StoreLocal(dTmp);			(* (top) dst...		*)
	e.PushValue(act, fmT);			(* (top) src,dst...	*)
      END;
      IF act.kind = Xp.mkStr THEN
	out.StaticCall(Vos.aaStrCopy, -2);	(* (top) ...		*)
      ELSE
	e.RefArrCopy(fmT);			(* (top) ...		*)
      END;
      out.LoadLocal(dTmp);			(* (top) dst...		*)
      out.proc.ReleaseLocal(dTmp);
      out.proc.ReleaseLocal(sTmp);
    END;
  END ValueCopy;

(* ---------------------------------------------------- *)

  PROCEDURE (e : LightningEmitter)StringCopy(act : D.Expr; fmT : Ty.Array),NEW;
    VAR out : Vos.IlasmFile;
  BEGIN
    out := e.outF;
    IF act.kind = Xp.mkStr THEN
      e.ValueCopy(act, fmT);
    ELSIF fmT.length = 0 THEN 		(* str passed to open array 	*)
      e.PushValue(act, fmT);
      e.outF.StaticCall(Vos.vStr2ChO, 0);
    ELSE				(* str passed to fixed array	*)
      out.PushInt(fmT.length);
      out.MkOpenArray(Ty.mkArrayOf(G.charTp));	
      out.Code(Asm.opc_dup);
      e.PushValue(act, fmT); 
      e.outF.StaticCall(Vos.vStr2ChF, -2);
    END;
  END StringCopy;

(* ============================================================ *)

  PROCEDURE (e : LightningEmitter)Invoke(exp : D.Expr; typ : Ty.Procedure),NEW;
    VAR code : INTEGER;
	prcI : Id.PrcId;
	mthI : Id.MthId;
  BEGIN
    WITH exp : Xp.IdLeaf DO (* qualid *)
	code := Asm.opc_call;
	prcI := exp.ident(Id.PrcId);
	IF prcI.kind # Id.ctorP THEN
	  e.outF.CallIT(code, prcI, typ);
	ELSE
	  e.outF.CallCT(prcI, typ);
	END;
    | exp : Xp.IdentX DO (* selct *)
	mthI := exp.ident(Id.MthId);
	IF exp.kind = Xp.sprMrk THEN 
	  code := Asm.opc_call;
	ELSIF mthI.bndType.isInterfaceType() THEN
	  code := Asm.opc_callvirt;
	ELSIF mthI.mthAtt * Id.mask = Id.final THEN
	  code := Asm.opc_call;
	ELSE
	  code := Asm.opc_callvirt;
        END;
	e.outF.CallIT(code, mthI, typ);
	IF Id.covar IN mthI.mthAtt THEN
	  e.outF.CodeT(Asm.opc_castclass, typ.retType);
	END;
    END;
  END Invoke;

(* ---------------------------------------------------- *)

  PROCEDURE (e : LightningEmitter)PushArg(act : D.Expr;
				          frm : Id.ParId),NEW;
    VAR idExp : Xp.IdentX;
	out   : Vos.IlasmFile;
	local : INTEGER;
  BEGIN

(* SPECIAL CASE OF VAR RECEIVER? *)

    out := e.outF;
    IF frm.boxOrd # D.val THEN
      e.PushRef(act, frm.type);
    ELSIF (act.type = G.strTp) & (frm.type IS Ty.Array) THEN (* a string *)
      e.StringCopy(act, frm.type(Ty.Array));
    ELSIF (frm.parMod = D.val) & ~Vos.hasValueRep(frm.type) THEN
      e.ValueCopy(act, frm.type);
    ELSE
      e.PushValue(act, frm.type);
    END;
  END PushArg;

(* ============================================================ *)
(*    Possible structures of procedure call expressions are:	*)
(* ============================================================ *)
(*	    o				    o			*)
(*   	   /			   	   /			*)
(*	[CallX] 			[CallX] 		*)
(*	 / +--- actuals --> ...          / +--- actuals 	*)
(*	/				/			*)
(*    [IdentX]			    [IdLeaf]			*)
(*      /  +--- ident ---> [Procs]      +--- ident ---> [Procs]	*)
(*     /							*)
(* kid expr							*)
(*								*)
(* ============================================================ *)
(*  only the right hand case can be a standard proc or function	*)
(* ============================================================ *)

  PROCEDURE (e : LightningEmitter)PushCall(callX : Xp.CallX),NEW;
    VAR jFile : Vos.IlasmFile;
	mark0 : SET;		(* local ord limit on entry *)
	tide0 : INTEGER;	(* parameter tide on entry  *)
	index : INTEGER;	(* just a counter for loops *)
	formT : Ty.Procedure;	(* formal type of procedure *)
	formP : Id.ParId;	(* current formal parameter *)
	prExp : D.Expr;
	idExp : Xp.IdentX;
 (* ---------------------------------------------------- *)
    PROCEDURE CheckCall(expr : D.Expr);
      VAR prcI : Id.PrcId;
	  mthI : Id.MthId;
    BEGIN
      WITH expr : Xp.IdLeaf DO (* qualid *)
	  prcI := expr.ident(Id.PrcId);
	  IF prcI.type.xName = NIL THEN Vos.MkCallAttr(prcI) END;
	  expr.type := prcI.type;
      | expr : Xp.IdentX DO (* selct *)
	  mthI := expr.ident(Id.MthId);
	  IF mthI.type.xName = NIL THEN Vos.MkCallAttr(mthI) END;
	  expr.type := mthI.type;
      END;
    END CheckCall;
 (* ---------------------------------------------------- *)
  BEGIN
    jFile := e.outF;
    mark0 := jFile.proc.markTop();
    tide0 := callX.actuals.tide;
    prExp := callX.kid;
   (*
    *  Before we push any arguments, we must make 
    *  sure that the formal-type name is computed.
    *)
    CheckCall(prExp);
    formT := prExp.type(Ty.Procedure);
   (*
    *  We must first deal with the receiver if this is a method.
    *)
    IF formT.receiver # NIL THEN
      jFile.CommentT("Start dispatch sequence");
      idExp := prExp(Xp.IdentX);
      formP := idExp.ident(Id.MthId).rcvFrm;
      e.PushArg(idExp.kid, formP);
    ELSE
      jFile.CommentT("Start static call sequence");
    END;
   (*
    *  We push the arguments from left to right.
    *)
    FOR index := 0 TO tide0-1 DO
      formP := formT.formals.a[index];
      e.PushArg(callX.actuals.a[index], formP);
    END;
   (*
    *  Now emit the actual call instruction(s)
    *)
    e.Invoke(prExp, formT);
   (*
    *  Now we save any out arguments from the appended exprs.
    *)
(*
#####
MIGHT NEED TO SAVE THE VAR RECEIVER HERE...
 *      e.CopyOut(idExp.kid, idExp.ident);
 *)
    jFile.CommentT("End call sequence");
    jFile.proc.ReleaseAll(mark0);
  END PushCall;

(* ---------------------------------------------------- *)

  PROCEDURE (e : LightningEmitter)EmitStdProc(callX : Xp.CallX),NEW;
    CONST fMsg = "Assertion failure ";
    VAR out  : Vos.IlasmFile;
	cTmp : INTEGER;
	rTmp : INTEGER;
	prId : Id.PrcId;
	vrId : Id.FldId;
	pOrd : INTEGER;
        arg0 : D.Expr;
	argX : D.Expr;
	dstT : D.Type;
	argN : INTEGER;
	numL : INTEGER;
	incr : INTEGER;
	okLb : INTEGER;
	long : BOOLEAN;
	ovfl : BOOLEAN;
	c    : INTEGER;
   (* --------------------------- *)
  BEGIN
    out  := e.outF;
    prId := callX.kid(Xp.IdLeaf).ident(Id.PrcId);
    arg0 := callX.actuals.a[0];	(* Always need at least one arg *)
    argN := callX.actuals.tide;
    pOrd := prId.stdOrd;
    CASE pOrd OF
   (* --------------------------- *)
    | G.asrtP :
	okLb := out.newLabel();
	e.FallFalse(arg0, okLb);
       (*
	*   If expression evaluates to false, fall
	*   into the error code, else skip to okLb.
	*)
	IF argN > 1 THEN
	  numL := intValue(callX.actuals.a[1]);
	  out.Trap(fMsg + LitValue.intToCharOpen(numL)^);
	ELSE
	  numL := callX.token.lin;
	  out.Trap(fMsg + CompState.srcNam +":"+ LitValue.intToCharOpen(numL)^);
	END;
	out.DefLab(okLb);
   (* --------------------------- *)
    | G.incP, G.decP :
	argX := callX.actuals.a[1];
	dstT := arg0.type;
        ovfl := out.proc.prId.ovfChk;
	e.PushHandle(arg0, dstT);
        WITH arg0 : Xp.IdLeaf DO
	    e.PushValue(arg0, dstT);
	| arg0 : Xp.IdentX DO
	    vrId := arg0.ident(Id.FldId);
	    out.Code(Asm.opc_dup);
	    out.PutGetF(Asm.opc_ldfld, vrId);
        | arg0 : Xp.BinaryX DO
(*
 *  Here is the decision point: the stack is currently
 *	(top) index,arrRef,...
 *  we might reduce the 2-slot handle to an address, then go
 *	dup; ldind; <add-n>; stind
 *  but would this verify?
 *  Alternatively, we can wimp out, and store both the ref
 *  and the index then go
 *      ldloc ref; ldloc idx; ldloc ref; ldloc idx; ldelem; <add-n>; stelem
 *)
	    rTmp := out.proc.newLocal(arg0.lKid.type);
	    cTmp := out.proc.newLocal(G.intTp);
	    out.StoreLocal(cTmp);	(* (top) ref,...		*)
	    out.Code(Asm.opc_dup);	(* (top) ref,ref,...		*)
	    out.StoreLocal(rTmp);	(* (top) ref,...		*)
	    out.LoadLocal(cTmp);	(* (top) idx,ref,...		*)
	    out.LoadLocal(rTmp);	(* (top) ref,idx,ref,...	*)
	    out.LoadLocal(cTmp);	(* (top) idx,ref,idx,ref,...	*)
	    out.GetElem(dstT);		(* (top) idx,ref,...		*)
	    out.proc.ReleaseLocal(cTmp);
	    out.proc.ReleaseLocal(rTmp);
	END;
	e.PushValue(argX, dstT);
	IF pOrd = G.incP THEN 
	  IF ovfl THEN c := Asm.opc_add_ovf ELSE c := Asm.opc_add END;
	ELSE 
	  IF ovfl THEN c := Asm.opc_sub_ovf ELSE c := Asm.opc_sub END;
	END;
	out.Code(c);
	e.ValueAssign(arg0);
   (* --------------------------- *)
    | G.exclP, G.inclP :
	dstT := arg0.type;
	argX := callX.actuals.a[1];
	e.PushHandle(arg0, dstT);
        IF arg0 IS Xp.IdLeaf THEN
	  e.PushValue(arg0, dstT);
	ELSE
	  WITH arg0 : Xp.BinaryX DO
	      ASSERT(arg0.kind = Xp.index);
	      rTmp := out.proc.newLocal(arg0.lKid.type);
	      cTmp := out.proc.newLocal(G.intTp);
	      out.StoreLocal(cTmp);
	      out.Code(Asm.opc_dup);
	      out.StoreLocal(rTmp);
	      out.LoadLocal(cTmp);	(* (top) idx,ref,...		*)
	      out.LoadLocal(rTmp);	(* (top) ref,idx,ref,...	*)
	      out.LoadLocal(cTmp);	(* (top) idx,ref,idx,ref,...	*)
	      out.GetElem(dstT);	(* (top) idx,ref,...		*)
	      out.proc.ReleaseLocal(cTmp);
	      out.proc.ReleaseLocal(rTmp);
	  | arg0 : Xp.IdentX DO
	      ASSERT(arg0.kind = Xp.selct);
	      out.Code(Asm.opc_dup);
	      out.PutGetF(Asm.opc_ldfld, arg0.ident(Id.FldId));
	  END;
	END;
	IF argX.kind = Xp.numLt THEN
	  out.PushInt(ORD({intValue(argX)}));
	ELSE
	  out.Code(Asm.opc_ldc_i4_1);
	  e.PushValue(argX, G.intTp);
	  out.Code(Asm.opc_shl);
	END;
	IF pOrd = G.inclP THEN
	  out.Code(Asm.opc_or);
	ELSE
	  out.Code(Asm.opc_ldc_i4_M1);
	  out.Code(Asm.opc_xor);
	  out.Code(Asm.opc_and);
	END;
	e.ValueAssign(arg0);
   (* --------------------------- *)
    | G.haltP :
	out.PushInt(intValue(arg0));
	out.StaticCall(Vos.sysExit, -1);
       (*
	*  We now do a dummy return to signal
	*  the verifier that control exits here.
	*)
	out.PushJunkAndQuit(out.proc.prId);
   (* --------------------------- *)
    | G.throwP :
	IF CompState.ntvExc.assignCompat(arg0) THEN
	  e.PushValue(arg0, CompState.ntvExc);
	  out.Code(Asm.opc_throw);
	ELSE
	  e.PushValue(arg0, CompState.ntvStr);
	  out.Throw();
	END;
   (* --------------------------- *)
    | G.newP :
       (*
	*   arg0 is a pointer to a Record or Array type.
	*)
	dstT := arg0.type(Ty.Pointer).boundTp;
	e.PushHandle(arg0, arg0.type);
        IF argN = 1 THEN
	 (*
	  *  No LEN argument implies either:
	  *	pointer to record, OR
	  *	pointer to a fixed array.
	  *)
	  WITH dstT : Ty.Record DO
	      out.MkNewRecord(dstT);
	  | dstT : Ty.Array DO
	      out.MkFixedArray(dstT);
	  END;
	ELSE
	  FOR numL := argN-1 TO 1 BY -1 DO
	    argX := callX.actuals.a[numL];
	    e.PushValue(argX, G.intTp);
	  END;
	  out.MkOpenArray(dstT(Ty.Array));
	END;
	e.ValueAssign(arg0);
   (* --------------------------- *)
    END;
  END EmitStdProc;

(* ============================================================ *)
(*		    Statement Handling Methods			*)
(* ============================================================ *)

  PROCEDURE (e : LightningEmitter)EmitAssign(stat : St.Assign),NEW;
    VAR lhTyp : D.Type;
  BEGIN
   (*
    *  This is a value assign in CP.
    *)
    lhTyp := stat.lhsX.type;
    IF Vos.hasValueRep(lhTyp) THEN
      e.PushHandle(stat.lhsX, lhTyp);
      e.PushValue(stat.rhsX, lhTyp);
      e.ValueAssign(stat.lhsX);
    ELSE (* a reference type *)
      e.PushValue(stat.lhsX, lhTyp);
      e.PushValue(stat.rhsX, lhTyp);
      WITH lhTyp : Ty.Array DO 
	IF stat.rhsX.type = G.strTp THEN
          e.outF.StaticCall(Vos.vStr2ChF, -2);
	ELSIF stat.rhsX.kind = Xp.mkStr THEN
	  e.outF.StaticCall(Vos.aaStrCopy, -2);
	ELSE
	  e.RefArrCopy(lhTyp);
	END;
      | lhTyp : Ty.Record DO
	  e.RefRecCopy(lhTyp);
      END;
    END;
  END EmitAssign;

(* ---------------------------------------------------- *)

  PROCEDURE (e : LightningEmitter)EmitCall(stat : St.ProcCall),NEW;
    VAR expr : Xp.CallX;	(* the stat call expression *)
  BEGIN
    expr := stat.expr(Xp.CallX);
    IF (expr.kind = Xp.prCall) & expr.kid.isStdProc() THEN
      e.EmitStdProc(expr);
    ELSE
      e.PushCall(expr);
    END;
  END EmitCall;

(* ---------------------------------------------------- *)

  PROCEDURE (e : LightningEmitter)
			EmitIf(stat : St.Choice; OUT ok : BOOLEAN),NEW;
    VAR out  : Vos.IlasmFile;
	high : INTEGER;			(* Branch count.  *)
	exLb : INTEGER;			(* Exit label	  *)
	nxtP : INTEGER;			(* Next predicate *)
	indx : INTEGER;
	live : BOOLEAN;			(* then is live   *)
	else : BOOLEAN;			(* else not seen  *)
	then : D.Stmt;
	pred : D.Expr;
  BEGIN
    ok := FALSE;
    out := e.outF;
    exLb := out.newLabel();
    else := FALSE;
    high := stat.preds.tide - 1;
    FOR indx := 0 TO high DO
      live := TRUE;
      pred := stat.preds.a[indx];
      then := stat.blocks.a[indx];
      nxtP := out.newLabel();
      IF pred = NIL THEN else := TRUE ELSE e.FallTrue(pred, nxtP) END;
      IF then # NIL THEN e.EmitStat(then, live) END;
      IF live THEN 
	ok := TRUE;
	IF indx < high THEN out.CodeLb(Asm.opc_br, exLb) END;
      END;
      out.DefLab(nxtP);
    END;
   (*
    *   If not ELSE has been seen, then control flow is still live!
    *)
    IF ~else THEN ok := TRUE END;
    out.DefLab(exLb);
  END EmitIf;

(* ---------------------------------------------------- *)

  PROCEDURE (e : LightningEmitter)EmitRanges
		       (locV : INTEGER;   	(* select Var   *)
			stat : St.CaseSt;  	(* case stat    *)
			minR : INTEGER;  	(* min rng-ix   *)
			maxR : INTEGER;  	(* max rng-ix   *)
			minI : INTEGER;  	(* min index    *)
			maxI : INTEGER;  	(* max index    *)
			dfLb : INTEGER),NEW;	(* default Lb   *)
   (* --------------------------------------------------------- * 
    *   This procedure emits the code for a single,
    *   dense range of selector values in the label-list.
    * --------------------------------------------------------- *)
    VAR out  : Vos.IlasmFile;
	loIx : INTEGER;		(* low selector value for dense range  *)
	hiIx : INTEGER;		(* high selector value for dense range *)
	rNum : INTEGER;		(* total number of ranges in the group *)
	peel : INTEGER;		(* max index of range to be peeled off *)
	indx : INTEGER;
	rnge : St.Triple;
  BEGIN
    out := e.outF;
    rNum := maxR - minR + 1;
    rnge := stat.labels.a[minR];
    IF rNum = 1 THEN		(* single range only *)
      out.EmitOneRange(locV, rnge.loC, rnge.hiC, rnge.ord, minI, maxI, dfLb);
    ELSIF rNum < 4 THEN	
     (*
      *    Two or three ranges only.
      *    Peel off the lowest of the ranges, and recurse.
      *)
      loIx := rnge.loC;
      peel := rnge.hiC;
      out.LoadLocal(locV);
     (*
      *   There are a number of special cases
      *   that can benefit from special code.
      *)
      IF loIx = peel THEN 
       (* 
        *   A singleton.  Leave minI unchanged, unless peel = minI.
        *)
	out.PushInt(peel);
	out.CodeLb(Asm.opc_beq, dfLb + rnge.ord + 1);
        IF minI = peel THEN minI := peel+1 END;
	INC(minR);
      ELSIF loIx = minI THEN
       (* 
        *   A range starting at the minimum selector value.
        *)
	out.PushInt(peel);
	out.CodeLb(Asm.opc_ble, dfLb + rnge.ord + 1);
        minI := peel+1;
        INC(minR);
      ELSE
       (* 
        *   We must peel the default range from minI to loIx.
        *)
	out.PushInt(loIx);
        out.CodeLb(Asm.opc_blt, dfLb);
        minI := loIx;	(* and minR is unchanged! *)
      END;
      e.EmitRanges(locV, stat, minR, maxR, minI, maxI, dfLb);
    ELSE 
     (*
      *   Four or more ranges.  Emit a dispatch table.
      *)
      loIx := rnge.loC;			(* low of min-range  *)
      hiIx := stat.labels.a[maxR].hiC;	(* high of max-range *)
      out.LoadLocal(locV);
      IF loIx # 0 THEN
	out.PushInt(loIx);
	out.Code(Asm.opc_sub);
      END;
      out.SwitchHead();
      (* ---- *)
      FOR indx := minR TO maxR DO
        rnge := stat.labels.a[indx];
        WHILE loIx < rnge.loC DO 
	  out.LstLab(dfLb); INC(loIx);
	END;
        WHILE loIx <= rnge.hiC DO 
	  out.LstLab(dfLb+rnge.ord+1); INC(loIx);
	END;
      END;
      (* ---- *)
      out.SwitchTail();
      out.CodeLb(Asm.opc_br, dfLb)
    END;
  END EmitRanges;

(* ---------------------------------------------------- *)

  PROCEDURE (e : LightningEmitter)EmitGroups
		       (locV : INTEGER;		(* select vOrd  *)
			stat : St.CaseSt;  	(* case stat  	*)
			minG : INTEGER;  	(* min grp-indx	*)
			maxG : INTEGER;  	(* max grp-indx	*)
			minI : INTEGER;  	(* min index  	*)
			maxI : INTEGER;  	(* max index  	*)
			dfLb : INTEGER),NEW;	(* default lab	*)
   (* --------------------------------------------------------- * 
    *  This function emits the branching code which sits on top
    *  of the selection code for each dense range of case values.
    * --------------------------------------------------------- *)
    VAR out   : Vos.IlasmFile;
	newLb : INTEGER;
	midPt : INTEGER;
	group : St.Triple;
	range : St.Triple;
  BEGIN
    IF minG = maxG THEN (* only one remaining dense group *)
      group := stat.groups.a[minG];
      e.EmitRanges(locV, stat, group.loC, group.hiC, minI, maxI, dfLb);
    ELSE
     (*
      *   We must bifurcate the group range, and recurse.
      *   We will split the value range at the lower limit 
      *   of the low-range of the upper half-group.
      *)
      midPt := (minG + maxG + 1) DIV 2;
      group := stat.groups.a[midPt];
      range := stat.labels.a[group.loC];
     (*
      *  Test and branch at range.loC
      *)
      out := e.outF;
      newLb := out.newLabel();
      out.LoadLocal(locV);
      out.PushInt(range.loC);
      out.CodeLb(Asm.opc_bge, newLb);
     (*
      *    Recurse!
      *)
      e.EmitGroups(locV, stat, minG, midPt-1, minI, range.loC-1, dfLb);
      out.DefLab(newLb);
      e.EmitGroups(locV, stat, midPt, maxG, range.loC, maxI, dfLb);
    END;
  END EmitGroups;

(* ---------------------------------------------------- *)

  PROCEDURE (e : LightningEmitter)
			EmitCase(stat : St.CaseSt; OUT ok : BOOLEAN),NEW;
    VAR out  : Vos.IlasmFile;
	indx : INTEGER;
	dfLb : INTEGER;
	exLb : INTEGER;
	selV : INTEGER;
	live : BOOLEAN;
	minI : INTEGER;
	maxI : INTEGER;
  BEGIN
   (* ---------------------------------------------------------- *
    *  CaseSt* = POINTER TO RECORD (D.Stmt)
    *	     (* ----------------------------------------- *
    *	      *	kind-  : INTEGER;	(* tag for unions *)
    *	      *	token* : S.Token;	(* stmt first tok *)
    *	      * ----------------------------------------- *)
    *		select* : D.Expr;	(* case selector  *)
    *		chrSel* : BOOLEAN;	(* ==> use chars  *)
    *		blocks* : D.StmtSeq;	(* case bodies    *)
    *		elsBlk* : D.Stmt;	(* elseCase | NIL *)
    *		labels* : TripleSeq;	(* label seqence  *)
    *		groups- : TripleSeq;	(* dense groups   *)
    *	      END;
    * --------------------------------------------------------- *
    *  Notes on the semantics of this structure. "blocks" holds	*
    *  an ordered list of case statement code blocks. "labels"	*
    *  is a list of ranges, intially in textual order,with flds	*
    *  loC, hiC and ord corresponding to the range min, max and	*
    *  the selected block ordinal number.  This list is later 	*
    *  sorted on the loC value, and adjacent values merged if 	*
    *  they select the same block. The "groups" list of triples *
    *  groups ranges into dense subranges in the selector space	*
    *  The fields loC, hiC, and ord to hold the lower and upper	*
    *  indices into the labels list, and the number of non-	*
    *  default values in the group. Groups are guaranteed to	*
    *  have density (nonDefN / (max-min+1)) > DENSITY		*
    * --------------------------------------------------------- *)
    ok := FALSE;
    out := e.outF;
    exLb := out.newLabel();
    dfLb := out.getLabelRange(stat.blocks.tide+1);
    IF stat.chrSel THEN 
      minI := 0; maxI := ORD(MAX(CHAR));
      selV := out.proc.newLocal(G.charTp);
    ELSE 
      minI := MIN(INTEGER); 
      maxI := MAX(INTEGER);
      selV := out.proc.newLocal(G.intTp);
    END;

   (*
    *    Push the selector value, and save in local variable;
    *)
    IF stat.chrSel THEN
      e.PushValue(stat.select, G.charTp);
    ELSE
      e.PushValue(stat.select, G.intTp);
    END;
    out.StoreLocal(selV);
    e.EmitGroups(selV, stat, 0, stat.groups.tide-1, minI, maxI, dfLb);
   (*
    *    Now we emit the code for the cases.
    *    If any branch returns, then exLb is reachable.
    *)
    FOR indx := 0 TO stat.blocks.tide-1 DO
      out.DefLab(dfLb + indx + 1);
      e.EmitStat(stat.blocks.a[indx], live);
      IF live THEN
	ok := TRUE;
        out.CodeLb(Asm.opc_br, exLb);
      END;
    END;
   (*
    *    Now we emit the code for the elespart.
    *    If the elsepart returns then exLb is reachable.
    *)
    out.DefLabC(dfLb, "Default case");
    IF stat.elsBlk # NIL THEN
      e.EmitStat(stat.elsBlk, live);
      IF live THEN ok := TRUE END;
    ELSE
      out.CaseTrap(selV);
    END;
    out.proc.ReleaseLocal(selV);
    IF ok THEN out.DefLabC(exLb, "Case exit label") END;
  END EmitCase;
 
(* ---------------------------------------------------- *)

  PROCEDURE (e : LightningEmitter)
			EmitWhile(stat : St.TestLoop; OUT ok : BOOLEAN),NEW;
    VAR out  : Vos.IlasmFile;
	lpLb : INTEGER;
	exLb : INTEGER;
  BEGIN
    out := e.outF;
    lpLb := out.newLabel();
    exLb := out.newLabel();
   (*
    *  Up to the Beta release of .NET encoding
    *  WHILE loops this way required the use of
    *  ilasm /debug to step around a JIT bug.
    *  Laying the WHILE out conventionally avoids 
    *  the bug, but costs at runtime.
    *)
    e.FallTrue(stat.test, exLb);	(* goto exLb if eval false *)
    out.DefLabC(lpLb, "Loop header");
    e.EmitStat(stat.body, ok);
    IF ok THEN e.FallFalse(stat.test, lpLb) END;
   (*
    *  Here is the alternative code ...
    *)
(*
 *  out.DefLabC(lpLb, "Loop header");
 *  e.FallTrue(stat.test, exLb);	(* goto exLb if eval false *)
 *  e.EmitStat(stat.body, ok);
 *  IF ok THEN out.CodeLb(Asm.opc_br, lpLb) END;
 *)
    out.DefLabC(exLb, "Loop exit");
  END EmitWhile;

(* ---------------------------------------------------- *)

  PROCEDURE (e : LightningEmitter)
			EmitRepeat(stat : St.TestLoop; OUT ok : BOOLEAN),NEW;
    VAR out  : Vos.IlasmFile;
	lpLb : INTEGER;
  BEGIN
    out := e.outF;
    lpLb := out.newLabel();
    out.DefLabC(lpLb, "Loop header");
    e.EmitStat(stat.body, ok);
    IF ok THEN e.FallTrue(stat.test, lpLb) END; (* exit on eval true *)
    out.CommentT("Loop exit");
  END EmitRepeat;

(* ---------------------------------------------------- *)

  PROCEDURE (e : LightningEmitter)
			EmitFor(stat : St.ForLoop; OUT ok : BOOLEAN),NEW;
   (* ----------------------------------------------------------- *
    *   This code has been split into the four cases:
    *   -	long control variable, counting up;
    *   -	long control variable, counting down;
    *   -	int control variable, counting up;
    *   -	int control variable, counting down;
    *   Of course, it is possible to fold all of this, and have 
    *   tests everywhere, but the following is cleaner, and easier 
    *   to enhance in the future.
    *
    *   Note carefully the use of ForLoop::isSimple().  It is 
    *   essential to use exactly the same function here as is
    *   used by ForLoop::flowAttr() for initialization analysis.
    *   If this were not the case, the verifier could barf.
    * ----------------------------------------------------------- *)
    PROCEDURE LongForUp(e: LightningEmitter; stat: St.ForLoop; OUT ok: BOOLEAN);
      VAR out  : Vos.IlasmFile;
	  cVar : Id.AbVar;
	  top1 : INTEGER;
	  exLb : INTEGER;
	  lpLb : INTEGER;
	  step : LONGINT;
	  smpl : BOOLEAN;
    BEGIN
      out := e.outF;
      lpLb := out.newLabel();
      exLb := out.newLabel();
      cVar := stat.cVar(Id.AbVar);
      step := longValue(stat.byXp); 
      smpl := stat.isSimple();
      IF smpl THEN
        out.PushLong(longValue(stat.loXp));
        out.PutVar(cVar);
        top1 := -1;			(* keep the verifier happy! *)
      ELSE
        top1 := out.proc.newLocal(G.lIntTp);
        e.PushValue(stat.hiXp, G.lIntTp);
        out.Code(Asm.opc_dup);
        out.StoreLocal(top1);
        e.PushValue(stat.loXp, G.lIntTp);
        out.Code(Asm.opc_dup);
        out.PutVar(cVar);
       (*
        *   The top test is NEVER inside the loop.
        *)
        e.DoCmp(Xp.lessT, exLb, G.lIntTp);
      END;
      out.DefLabC(lpLb, "Loop header");
     (*
      *   Emit the code body.
      *   Stack contents are (top) hi, ...
      *   and exactly the same on the backedge.
      *)
      e.EmitStat(stat.body, ok);
     (*
      *   If the body returns ... do an exit test.
      *)
      IF ok THEN 
        IF smpl THEN
          out.PushLong(longValue(stat.hiXp));
	ELSE
          out.LoadLocal(top1);
	END;
        out.GetVar(cVar);			(* (top) cv,hi		*)
	out.PushLong(step);
        out.Code(Asm.opc_add_ovf);		(* (top) cv',hi		*)
        out.Code(Asm.opc_dup);			(* (top) cv',cv',hi	*)
        out.PutVar(cVar);			(* (top) cv',hi		*)
        e.DoCmp(Xp.greEq,  lpLb, G.lIntTp);
      END;
     (*
      *   The exit label.
      *)
      out.DefLabC(exLb, "Loop trailer");
    END LongForUp;

   (* ----------------------------------------- *)

    PROCEDURE LongForDn(e: LightningEmitter; stat: St.ForLoop; OUT ok: BOOLEAN);
      VAR out  : Vos.IlasmFile;
	  cVar : Id.AbVar;
	  top1 : INTEGER;
	  exLb : INTEGER;
	  lpLb : INTEGER;
	  step : LONGINT;
	  smpl : BOOLEAN;
    BEGIN
      out := e.outF;
      lpLb := out.newLabel();
      exLb := out.newLabel();
      cVar := stat.cVar(Id.AbVar);
      step := longValue(stat.byXp);
      smpl := stat.isSimple();
      IF smpl THEN
        out.PushLong(longValue(stat.loXp));
        out.PutVar(cVar);
        top1 := -1;			(* keep the verifier happy! *)
      ELSE
        top1 := out.proc.newLocal(G.lIntTp);
        e.PushValue(stat.hiXp, G.lIntTp);
        out.Code(Asm.opc_dup);
        out.StoreLocal(top1);
        e.PushValue(stat.loXp, G.lIntTp);
        out.Code(Asm.opc_dup);
        out.PutVar(cVar);
       (*
        *   The top test is NEVER inside the loop.
        *)
        e.DoCmp(Xp.greT,  exLb, G.lIntTp);
      END;
      out.DefLabC(lpLb, "Loop header");
     (*
      *   Emit the code body.
      *   Stack contents are (top) hi, ...
      *   and exactly the same on the backedge.
      *)
      e.EmitStat(stat.body, ok);
     (*
      *   If the body returns ... do an exit test.
      *)
      IF ok THEN 
        IF smpl THEN
          out.PushLong(longValue(stat.hiXp));
	ELSE
          out.LoadLocal(top1);
	END;
        out.GetVar(cVar);			(* (top) cv,hi		*)
	out.PushLong(step);
        out.Code(Asm.opc_add_ovf);		(* (top) cv',hi		*)
        out.Code(Asm.opc_dup);			(* (top) cv',cv',hi	*)
        out.PutVar(cVar);			(* (top) cv',hi		*)
        e.DoCmp(Xp.lessEq, lpLb, G.lIntTp);
      END;
     (*
      *   The exit label.
      *)
      out.DefLabC(exLb, "Loop trailer");
    END LongForDn;

   (* ----------------------------------------- *)

    PROCEDURE IntForUp(e: LightningEmitter; stat: St.ForLoop; OUT ok: BOOLEAN);
      VAR out  : Vos.IlasmFile;
	  cVar : Id.AbVar;
	  topV : INTEGER;
	  exLb : INTEGER;
	  lpLb : INTEGER;
	  step : INTEGER;
	  smpl : BOOLEAN;
    BEGIN
     (*
      *    This is the common case, so we work a bit harder.
      *)
      out := e.outF;
      lpLb := out.newLabel();
      exLb := out.newLabel();
      cVar := stat.cVar(Id.AbVar);
      step := intValue(stat.byXp);
      smpl := stat.isSimple();
      IF smpl THEN
        out.PushInt(intValue(stat.loXp));
        out.PutVar(cVar);
        topV := -1;			(* keep the verifier happy! *)
      ELSE
        topV := out.proc.newLocal(G.intTp);
        e.PushValue(stat.hiXp, G.intTp);
        out.Code(Asm.opc_dup);
        out.StoreLocal(topV);
        e.PushValue(stat.loXp, G.intTp);
        out.Code(Asm.opc_dup);
        out.PutVar(cVar);
       (*
        *   The top test is NEVER inside the loop.
        *)
        e.DoCmp(Xp.lessT, exLb, G.intTp);
      END;
      out.DefLabC(lpLb, "Loop header");
     (*
      *   Emit the code body.
      *)
      e.EmitStat(stat.body, ok);
     (*
      *   If the body returns ... do an exit test.
      *)
      IF ok THEN 
	IF smpl THEN
	  out.PushInt(intValue(stat.hiXp));
	ELSE
          out.LoadLocal(topV);
	END;
        out.GetVar(cVar);			(* (top) cv,hi		*)
	out.PushInt(step);
        out.Code(Asm.opc_add_ovf);		(* (top) cv',hi		*)
        out.Code(Asm.opc_dup);			(* (top) cv',cv',hi	*)
        out.PutVar(cVar);			(* (top) cv',hi		*)
        e.DoCmp(Xp.greEq, lpLb, G.intTp);
      END;
     (*
      *   The exit label.
      *)
      out.DefLabC(exLb, "Loop trailer");
    END IntForUp;
  
   (* ----------------------------------------- *)

    PROCEDURE IntForDn(e: LightningEmitter; stat: St.ForLoop; OUT ok: BOOLEAN);
      VAR out  : Vos.IlasmFile;
	  cVar : Id.AbVar;
	  topV : INTEGER;
	  exLb : INTEGER;
	  lpLb : INTEGER;
	  step : INTEGER;
	  smpl : BOOLEAN;
    BEGIN
      out := e.outF;
      lpLb := out.newLabel();
      exLb := out.newLabel();
      cVar := stat.cVar(Id.AbVar);
      step := intValue(stat.byXp);
      topV := out.proc.newLocal(G.intTp);
      smpl := stat.isSimple();
      IF smpl THEN
        out.PushInt(intValue(stat.loXp));
        out.PutVar(cVar);
        topV := -1;			(* keep the verifier happy! *)
      ELSE
        e.PushValue(stat.hiXp, G.intTp);
        out.Code(Asm.opc_dup);
        out.StoreLocal(topV);
        e.PushValue(stat.loXp, G.intTp);
        out.Code(Asm.opc_dup);
        out.PutVar(cVar);
       (*
        *   The top test is NEVER inside the loop.
        *)
        e.DoCmp(Xp.greT, exLb, G.intTp);
      END;
      out.DefLabC(lpLb, "Loop header");
     (*
      *   Emit the code body.
      *)
      e.EmitStat(stat.body, ok);
     (*
      *   If the body returns ... do an exit test.
      *)
      IF ok THEN 
	IF smpl THEN
          out.PushInt(intValue(stat.hiXp));
	ELSE
          out.LoadLocal(topV);
	END;
        out.GetVar(cVar);			(* (top) cv,hi		*)
	out.PushInt(step);
        out.Code(Asm.opc_add_ovf);		(* (top) cv',hi		*)
        out.Code(Asm.opc_dup);			(* (top) cv',cv',hi	*)
        out.PutVar(cVar);			(* (top) cv',hi		*)
        e.DoCmp(Xp.lessEq, lpLb, G.intTp);
      END;
     (*
      *   The exit label.
      *)
      out.DefLabC(exLb, "Loop trailer");
    END IntForDn;

   (* ----------------------------------------- *)
  BEGIN (* body of EmitFor *)
    IF stat.cVar.type.isLongType() THEN 
      IF longValue(stat.byXp) > 0 THEN LongForUp(e, stat, ok);
      ELSE LongForDn(e, stat, ok);
      END;
    ELSE
      IF longValue(stat.byXp) > 0 THEN IntForUp(e, stat, ok);
      ELSE IntForDn(e, stat, ok);
      END;
    END;
  END EmitFor;

(* ---------------------------------------------------- *)

  PROCEDURE (e : LightningEmitter)
			EmitLoop(stat : St.TestLoop; OUT ok : BOOLEAN),NEW;
    VAR out  : Vos.IlasmFile;
	lpLb : INTEGER;
  BEGIN
    out := e.outF;
    lpLb  := out.newLabel();
    stat.label := out.newLabel();
    out.DefLabC(lpLb, "Loop header");
    e.EmitStat(stat.body, ok);
    IF ok THEN out.CodeLb(Asm.opc_br, lpLb) END;
    out.DefLabC(stat.label, "Loop exit");
  END EmitLoop;

(* ---------------------------------------------------- *)

  PROCEDURE (e : LightningEmitter)
			EmitWith(stat : St.Choice; OUT ok : BOOLEAN),NEW;
    VAR out  : Vos.IlasmFile;
	high : INTEGER;			(* Branch count.  *)
	exLb : INTEGER;			(* Exit label	  *)
	nxtP : INTEGER;			(* Next predicate *)
	indx : INTEGER;
	live : BOOLEAN;
	then : D.Stmt;
	pred : Xp.BinaryX;
	tVar : Id.LocId;
   (* --------------------------- *)
    PROCEDURE WithTest(je : LightningEmitter; 
		       os : Vos.IlasmFile; 
		       pr : Xp.BinaryX;
		       nx : INTEGER;
		       to : INTEGER);
      VAR ty : D.Type;
    BEGIN
      ty := pr.rKid(Xp.IdLeaf).ident.type;
      je.PushValue(pr.lKid, pr.lKid.type);
      os.CodeT(Asm.opc_isinst, ty);
     (*
      *  isinst returns the cast type, or NIL
      *  We save this to the allocated temp or needed type.
      *)
      os.StoreLocal(to);
      os.LoadLocal(to);
      os.CodeLb(Asm.opc_brfalse, nx);  (* branch on NIL *)
    END WithTest;
   (* --------------------------- *)
  BEGIN
    tVar := NIL;
    pred := NIL;
    ok := FALSE;
    out := e.outF;
    exLb := out.newLabel();
    high := stat.preds.tide - 1;
    FOR indx := 0 TO high DO
      live := TRUE;
      then := stat.blocks.a[indx];
      pred := stat.preds.a[indx](Xp.BinaryX);
      tVar := stat.temps.a[indx](Id.LocId);
      nxtP := out.newLabel();
      IF pred # NIL THEN 
        tVar.varOrd := out.proc.newLocal(tVar.type);
	WithTest(e, out, pred, nxtP, tVar.varOrd);
      END;
      IF then # NIL THEN e.EmitStat(then, live) END;
      IF live THEN 
	ok := TRUE;
       (*
	*  If this is not the else case, skip over the
	*  later cases, or jump over the WITH ELSE trap.
	*)
	IF pred # NIL THEN out.CodeLb(Asm.opc_br, exLb) END;
      END;
      out.DefLab(nxtP);
      IF tVar # NIL THEN out.proc.ReleaseLocal(tVar.varOrd) END;
    END;
    IF pred # NIL THEN out.WithTrap(pred(Xp.BinaryX).lKid(Xp.IdLeaf).ident) END;
    out.DefLab(exLb);
  END EmitWith;

(* ---------------------------------------------------- *)

  PROCEDURE (e : LightningEmitter)EmitExit(stat : St.ExitSt),NEW;
  BEGIN
    e.outF.CodeLb(Asm.opc_br, stat.loop(St.TestLoop).label);
  END EmitExit;

(* ---------------------------------------------------- *)

  PROCEDURE (e : LightningEmitter)EmitReturn(stat : St.Return),NEW;
    VAR out : Vos.IlasmFile;
	pId : Id.Procs;
	ret : D.Expr;
  BEGIN
    out := e.outF;
    ret := stat.retX;
    IF stat.retX # NIL THEN e.PushValue(ret, ret.type) END;
    out.DoReturn;
  END EmitReturn;

(* ---------------------------------------------------- *)

  PROCEDURE (e : LightningEmitter)
			EmitBlock(stat : St.Block; OUT ok : BOOLEAN),NEW;
    VAR index, limit : INTEGER;
  BEGIN
    ok := TRUE;
    index := 0;
    limit := stat.sequ.tide;
    WHILE ok & (index < limit) DO
      e.EmitStat(stat.sequ.a[index], ok);
      INC(index);
    END;
  END EmitBlock;

(* ---------------------------------------------------- *)
(* ---------------------------------------------------- *)

  PROCEDURE (e : LightningEmitter)EmitStat(stat : D.Stmt; OUT ok : BOOLEAN),NEW;
    VAR depth : INTEGER;
	out   : Vos.IlasmFile;
  BEGIN
    IF (stat = NIL) OR (stat.kind = St.emptyS) THEN ok := TRUE; RETURN END;
    out := e.outF;
    IF stat.kind # St.blockS THEN 
      out.Line(stat.token.lin);
    END;
    depth := out.proc.getDepth();
    CASE stat.kind OF
    | St.assignS  : e.EmitAssign(stat(St.Assign)); ok := TRUE;
    | St.procCall : e.EmitCall(stat(St.ProcCall)); ok := TRUE; 
    | St.ifStat   : e.EmitIf(stat(St.Choice), ok);
    | St.caseS    : e.EmitCase(stat(St.CaseSt), ok); 
    | St.whileS   : e.EmitWhile(stat(St.TestLoop), ok);
    | St.repeatS  : e.EmitRepeat(stat(St.TestLoop), ok); 
    | St.forStat  : e.EmitFor(stat(St.ForLoop), ok); 
    | St.loopS    : e.EmitLoop(stat(St.TestLoop), ok);
    | St.withS    : e.EmitWith(stat(St.Choice), ok);
    | St.exitS    : e.EmitExit(stat(St.ExitSt)); ok := TRUE;
    | St.returnS  : e.EmitReturn(stat(St.Return)); ok := FALSE;
    | St.blockS   : e.EmitBlock(stat(St.Block), ok);
    END;
    out.proc.SetDepth(depth);
  END EmitStat;

(* ============================================================ *)
(* ============================================================ *)
END LightningMaker.
(* ============================================================ *)
(* ============================================================ *)
