(* ==================================================================== *)
(*									*)
(*  SymFileRW:  Symbol-file reading and writing for GPCP.		*)
(*	Copyright (c) John Gough 1999, 2000.				*)
(*									*)
(* ==================================================================== *)

MODULE SymFileRW;

  IMPORT 
	GPCPcopyright,
	RTS,
	Error,
	Console,
	BF := GPBinFiles,
	Id := IdDesc,
	D  := Symbols,
	LitValue,
	Visitor,
	ExprDesc,
	Ty := TypeDesc,
	B  := Builtin,
	S  := CPascalS,
	G  := CompState,
	Nh := NameHash,
	FileNames;

(* ========================================================================= *
// Collected syntax ---
// 
// SymFile    = Header [String (falSy | truSy | <other attribute>)]
//		{Import | Constant | Variable | Type | Procedure} 
//		TypeList Key.
//	-- optional String is external name.
//	-- falSy ==> Java class
//	-- truSy ==> Java interface
//	-- others ...
// Header     = magic modSy Name.
// Import     = impSy Name [String] Key.
//	-- optional string is explicit external name of class
// Constant   = conSy Name Literal.
// Variable   = varSy Name TypeOrd.
// Type       = typSy Name TypeOrd.
// Procedure  = prcSy Name [String] FormalType.
//	-- optional string is explicit external name of procedure
// Method     = mthSy Name byte byte TypeOrd [String] FormalType.
//	-- optional string is explicit external name of method
// FormalType = [retSy TypeOrd] frmSy {parSy byte TypeOrd} endFm.
//	-- optional phrase is return type for proper procedures
// TypeOrd    = ordinal.
// TypeHeader = tDefS Ord [fromS Ord Name].
//	-- optional phrase occurs if:
//	-- type not from this module, i.e. indirect export
// TypeList   = start { Array | Record | Pointer | ProcType } close.
// Array      = TypeHeader arrSy TypeOrd (Byte | Number | <empty>) endAr.
//	-- nullable phrase is array length for fixed length arrays
// Pointer    = TypeHeader ptrSy TypeOrd.
// ProcType   = TypeHeader pTpSy FormalType.
// Record     = TypeHeader recSy recAtt [truSy | falSy] [basSy TypeOrd] 
//		{Name TypeOrd} {Method} {Statics} endRc.
//	-- truSy ==> is an extension of external interface
//	-- falSy ==> is an extension of external class
// 	-- basSy option defines base type, if not ANY / j.l.Object
// Statics    = ( Constant | Variable | Procedure ).
// Enum       = TypeHeader eTpSy { Constant } endRc.
// Name	      = namSy byte UTFstring.
// Literal    = Number | String | Set | Char | Real | falSy | truSy.
// Byte       = bytSy byte.
// String     = strSy UTFstring.
// Number     = numSy longint.
// Real       = fltSy ieee-double.
// Set        = setSy integer.
// Key        = keySy integer..
// Char       = chrSy unicode character.
//
// Notes on the syntax:
// All record types must have a Name field, even though this is often
// redundant.  The issue is that every record type (including those that
// are anonymous in CP) corresponds to a IR class, and the definer 
// and the user of the class _must_ agree on the IR name of the class.
// The same reasoning applies to procedure types, which must have equal
// interface names in all modules.
// ======================================================================== *)

  CONST
	modSy = ORD('H'); namSy = ORD('$'); bytSy = ORD('\');
	numSy = ORD('#'); chrSy = ORD('c'); strSy = ORD('s');
	fltSy = ORD('r'); falSy = ORD('0'); truSy = ORD('1');
	impSy = ORD('I'); setSy = ORD('S'); keySy = ORD('K');
	conSy = ORD('C'); typSy = ORD('T'); tDefS = ORD('t');
	prcSy = ORD('P'); retSy = ORD('R'); mthSy = ORD('M');
	varSy = ORD('V'); parSy = ORD('p'); start = ORD('&');
	close = ORD('!'); recSy = ORD('{'); endRc = ORD('}');
	frmSy = ORD('('); fromS = ORD('@'); endFm = ORD(')');
	arrSy = ORD('['); endAr = ORD(']'); pTpSy = ORD('%');
	ptrSy = ORD('^'); basSy = ORD('+'); eTpSy = ORD('e');

  CONST
	magic   = 0DEADD0D0H;
	syMag   = 0D0D0DEADH;
	dumped* = -1;

(* ============================================================ *)

  TYPE
	SymFile = POINTER TO RECORD 
		    file : BF.FILE;
		    cSum : INTEGER;
		    modS : Id.BlkId;
		    iNxt : INTEGER;
		    oNxt : INTEGER;
		    work : D.TypeSeq;
		  END;

  TYPE
	SymFileReader* = POINTER TO RECORD
		    file  : BF.FILE;
		    modS  : Id.BlkId;
		    impS  : Id.BlkId;
		    sSym  : INTEGER;
		    cAtt  : CHAR;
		    iAtt  : INTEGER;
		    lAtt  : LONGINT;
		    rAtt  : REAL;
		    sAtt  : FileNames.NameString;
		    sArray : D.ScpSeq;		(* These two sequences	*)
  		    tArray : D.TypeSeq;		(* must be private as   *)
		  END;				(* file parses overlap. *)

(* ============================================================ *)

  TYPE	TypeLinker* = POINTER TO RECORD (D.SymForAll) sym : SymFileReader END;
  TYPE	SymFileSFA* = POINTER TO RECORD (D.SymForAll) sym : SymFile END;

(* ============================================================ *)

  VAR   lastKeyVal : INTEGER;	(* private state for CPMake *)

(* ============================================================ *)
(* ========	     Import Stack Implementation	======= *)
(* ============================================================ *)

   VAR	stack	: ARRAY 32 OF Id.BlkId;
	topIx	: INTEGER;

   PROCEDURE InitStack;
   BEGIN
     topIx := 0; G.impMax := 0;
   END InitStack;

   PROCEDURE PushStack(b : Id.BlkId);
   BEGIN
     stack[topIx] := b; 
     INC(topIx);
     IF topIx > G.impMax THEN G.impMax := topIx END; 
   END PushStack;

   PROCEDURE PopStack;
   BEGIN
     DEC(topIx);
   END PopStack;

(* ============================================================ *)

  PROCEDURE GetLastKeyVal*() : INTEGER;
  BEGIN
    RETURN lastKeyVal;
  END GetLastKeyVal;

(* ============================================================ *)
(* ========	Various writing utility procedures	======= *)
(* ============================================================ *)

  PROCEDURE newSymFile(mod : Id.BlkId) : SymFile;
    VAR new : SymFile;
  BEGIN
    NEW(new);
   (*
    *  Initialization: cSum starts at zero. Since impOrd of
    *  the module is zero, impOrd of the imports starts at 1.
    *)
    new.cSum := 0;
    new.iNxt := 1;
    new.oNxt := D.tOffset;
    new.modS := mod;
    D.InitTypeSeq(new.work, 32);
    RETURN new;
  END newSymFile;

(* ======================================= *)

  PROCEDURE (f : SymFile)Write(chr : INTEGER),NEW;
    VAR tmp : INTEGER;
  BEGIN [UNCHECKED_ARITHMETIC]
   (* need to turn off overflow checking here *)
    tmp := f.cSum * 2 + chr;
    IF f.cSum < 0 THEN INC(tmp) END;
    f.cSum := tmp;
    BF.WriteByte(f.file, chr);
  END Write;

(* ======================================= *)

  PROCEDURE (f : SymFile)WriteStrUTF(IN nam : ARRAY OF CHAR),NEW;
    VAR buf : ARRAY 256 OF INTEGER;
	num : INTEGER;
	idx : INTEGER;
	chr : INTEGER;
  BEGIN
    num := 0;
    idx := 0;
    chr := ORD(nam[idx]);
    WHILE chr # 0H DO
      IF    chr <= 7FH THEN
	buf[num] := chr; INC(num);
      ELSIF chr <= 7FFH THEN
	buf[num+1] := 080H + chr MOD 64; chr := chr DIV 64;
	buf[num  ] := 0C0H + chr; INC(num, 2);
      ELSE
	buf[num+2] := 080H + chr MOD 64; chr := chr DIV 64;
	buf[num+1] := 080H + chr MOD 64; chr := chr DIV 64;
	buf[num  ] := 0E0H + chr; INC(num, 3);
      END;
      INC(idx); chr := ORD(nam[idx]);
    END;
    f.Write(num DIV 256);
    f.Write(num MOD 256);
    FOR idx := 0 TO num-1 DO f.Write(buf[idx]) END;
  END WriteStrUTF;

(* ======================================= *)

  PROCEDURE (f : SymFile)WriteOpenUTF(chOp : LitValue.CharOpen),NEW;
    VAR buf : ARRAY 256 OF INTEGER;
	num : INTEGER;
	idx : INTEGER;
	chr : INTEGER;
  BEGIN
    num := 0;
    idx := 0;
    chr := ORD(chOp[0]);
    WHILE chr # 0H DO
      IF    chr <= 7FH THEN
	buf[num] := chr; INC(num);
      ELSIF chr <= 7FFH THEN
	buf[num+1] := 080H + chr MOD 64; chr := chr DIV 64;
	buf[num  ] := 0C0H + chr; INC(num, 2);
      ELSE
	buf[num+2] := 080H + chr MOD 64; chr := chr DIV 64;
	buf[num+1] := 080H + chr MOD 64; chr := chr DIV 64;
	buf[num  ] := 0E0H + chr; INC(num, 3);
      END;
      INC(idx);
      chr := ORD(chOp[idx]);
    END;
    f.Write(num DIV 256);
    f.Write(num MOD 256);
    FOR idx := 0 TO num-1 DO f.Write(buf[idx]) END;
  END WriteOpenUTF;

(* ======================================= *)

  PROCEDURE (f : SymFile)WriteString(IN nam : ARRAY OF CHAR),NEW;
  BEGIN
    f.Write(strSy); 
    f.WriteStrUTF(nam);
  END WriteString;

(* ======================================= *)

  PROCEDURE (f : SymFile)WriteName(idD : D.Idnt),NEW;
  BEGIN
    f.Write(namSy); 
    f.Write(idD.vMod); 
    f.WriteOpenUTF(Nh.charOpenOfHash(idD.hash));
  END WriteName;

(* ======================================= *)

  PROCEDURE (f : SymFile)WriteChar(chr : CHAR),NEW;
    CONST mask = {0 .. 7};
    VAR   a,b,int : INTEGER;
  BEGIN
    f.Write(chrSy);
    int := ORD(chr);
    b := ORD(BITS(int) * mask); int := ASH(int, -8);
    a := ORD(BITS(int) * mask); 
    f.Write(a); 
    f.Write(b); 
  END WriteChar;

(* ======================================= *)

  PROCEDURE (f : SymFile)Write4B(int : INTEGER),NEW;
    CONST mask = {0 .. 7};
    VAR   a,b,c,d : INTEGER;
  BEGIN
    d := ORD(BITS(int) * mask); int := ASH(int, -8);
    c := ORD(BITS(int) * mask); int := ASH(int, -8);
    b := ORD(BITS(int) * mask); int := ASH(int, -8);
    a := ORD(BITS(int) * mask); 
    f.Write(a); 
    f.Write(b); 
    f.Write(c); 
    f.Write(d); 
  END Write4B;

(* ======================================= *)

  PROCEDURE (f : SymFile)Write8B(val : LONGINT),NEW;
  BEGIN
    f.Write4B(RTS.hiInt(val));
    f.Write4B(RTS.loInt(val));
  END Write8B;

(* ======================================= *)

  PROCEDURE (f : SymFile)WriteNum(val : LONGINT),NEW;
  BEGIN
    f.Write(numSy);
    f.Write8B(val);
  END WriteNum;

(* ======================================= *)

  PROCEDURE (f : SymFile)WriteReal(flt : REAL),NEW;
    VAR rslt : LONGINT;
  BEGIN
    f.Write(fltSy);
    rslt := RTS.realToLongBits(flt);
    f.Write8B(rslt);
  END WriteReal;

(* ======================================= *)

  PROCEDURE (f : SymFile)WriteOrd(ord : INTEGER),NEW;
  BEGIN
    IF ord <= 7FH THEN 
      f.Write(ord);
    ELSIF ord <= 7FFFH THEN
      f.Write(128 + ord MOD 128);	(* LS7-bits first *)
      f.Write(ord DIV 128);		(* MS8-bits next  *)
    ELSE
      ASSERT(FALSE);
    END;
  END WriteOrd;

(* ======================================= *)

  PROCEDURE (f : SymFile)EmitTypeOrd(t : D.Type),NEW;
  (*
   *  This proceedure facilitates the naming rules
   *  for records and (runtime) classes: -
   *
   *  (1)  Classes derived from named record types have
   *       names synthesized from the record typename.
   *  (2)  If a named pointer is bound to an anon record
   *       the class takes its name from the pointer name.
   *  (3)  If both the pointer and the record types have
   *       names, the class is named from the record.
   *)
    VAR recT : Ty.Record;
   (* ------------------------------------ *)
    PROCEDURE AddToWorklist(syF :SymFile; tyD : D.Type);
    BEGIN
      tyD.dump := syF.oNxt; INC(syF.oNxt);
      D.AppendType(syF.work, tyD);
      IF tyD.idnt = NIL THEN tyD.idnt := Id.newAnonId(tyD.dump) END;
    END AddToWorklist;
   (* ------------------------------------ *)
  BEGIN
    IF t.dump = 0 THEN (* type is not dumped yet *)
      WITH t : Ty.Record DO
       (*
        *   We wish to ensure that anonymous records are
        *   never emitted before their binding pointer
        *   types.  This ensures that we do not need to
        *   merge types when reading the files.
        *)
	IF (t.bindTp # NIL) & 
	   (t.bindTp.dump = 0) THEN 
	  AddToWorklist(f, t.bindTp);		(* First the pointer...  *)
	END;
        AddToWorklist(f, t);			(* Then this record type *)
      | t : Ty.Pointer DO
       (*
	*  If a pointer to record is being emitted, and 
        *  the pointer is NOT anonymous, then the class
	*  is known by the name of the record.  Thus the
	*  record name must be emitted, at least opaquely.
	*  Furthermore, we must indicate the binding
	*  relationship between the pointer and record.
	*  (It is possible that DCode need record size.)
	*)
        AddToWorklist(f, t);			(* First this pointer... *)
	IF (t.boundTp # NIL) & 
	   (t.boundTp.dump = 0) &
	   (t.boundTp IS Ty.Record) THEN
	  recT := t.boundTp(Ty.Record);
	  IF recT.bindTp = NIL THEN
            t.force := D.forced;
	    AddToWorklist(f, t.boundTp);	(* Then the record type  *)
	  END;
	END;
      ELSE (* All others *)
        AddToWorklist(f, t);			(* Just add the type.    *)
      END;
    END;
    f.WriteOrd(t.dump);
  END EmitTypeOrd;

(* ============================================================ *)
(* ========	    Various writing procedures		======= *)
(* ============================================================ *)

  PROCEDURE (f : SymFile)FormalType(t : Ty.Procedure),NEW;
  (*
  ** FormalType = [retSy TypeOrd] frmSy {parSy Byte TypeOrd} endFm.
  *)
    VAR indx : INTEGER;
	parI : Id.ParId;
  BEGIN
    IF t.retType # NIL THEN
      f.Write(retSy);
      f.EmitTypeOrd(t.retType);
     (*
      *   The structure of this type must be
      *   emitted, unless it is an imported type.
      *)
      t.retType.ConditionalMark();
    END;
    f.Write(frmSy);
    FOR indx := 0 TO t.formals.tide-1 DO
      parI := t.formals.a[indx];
      f.Write(parSy);
      f.Write(parI.parMod);
      f.EmitTypeOrd(parI.type);
     (*
      *   The structure of this type must be
      *   emitted, unless it is an imported type.
      *)
      parI.type.ConditionalMark();
    END;
    f.Write(endFm);
  END FormalType;

(* ======================================= *)

  PROCEDURE (f : SymFile)EmitConstId(id : Id.ConId),NEW;
    VAR conX : ExprDesc.LeafX;
        cVal : LitValue.Value;
	sVal : INTEGER;
  (*
  ** Constant = conSy Name Literal.
  ** Literal  = Number | String | Set | Char | Real | falSy | truSy.
  *)
  BEGIN
    conX := id.conExp(ExprDesc.LeafX);
    cVal := conX.value;
    f.Write(conSy);
    f.WriteName(id);
    CASE conX.kind OF
    | ExprDesc.tBool  : f.Write(truSy);
    | ExprDesc.fBool  : f.Write(falSy);
    | ExprDesc.numLt  : f.WriteNum(cVal.long());
    | ExprDesc.charLt : f.WriteChar(cVal.char());
    | ExprDesc.realLt : f.WriteReal(cVal.real());
    | ExprDesc.strLt  : f.WriteString(cVal.chOpen());
    | ExprDesc.setLt  : 
	f.Write(setSy); 
	IF cVal # NIL THEN sVal := cVal.int() ELSE sVal := 0 END;
	f.Write4B(sVal);
    END;
  END EmitConstId;

(* ======================================= *)

  PROCEDURE (f : SymFile)EmitTypeId(id : Id.TypId),NEW;
  (*
  **  Type = TypeSy Name TypeOrd.
  *)
  BEGIN
    f.Write(typSy);
    f.WriteName(id);
    f.EmitTypeOrd(id.type);
   (*
    *   The structure of this type must be
    *   emitted, even if it is an imported type.
    *)
    id.type.UnconditionalMark();
  END EmitTypeId;

(* ======================================= *)

  PROCEDURE (f : SymFile)EmitVariableId(id : Id.VarId),NEW;
  (*
  ** Variable = varSy Name TypeOrd.
  *)
  BEGIN
    f.Write(varSy);
    f.WriteName(id);
    f.EmitTypeOrd(id.type);
   (*
    *   The structure of this type must be
    *   emitted, unless it is an imported type.
    *)
    id.type.ConditionalMark();
  END EmitVariableId;

(* ======================================= *)

  PROCEDURE (f : SymFile)EmitImportId(id : Id.BlkId),NEW;
  (*
  ** Import = impSy Name.
  *)
  BEGIN
    IF D.need IN id.xAttr THEN
      f.Write(impSy);
      f.WriteName(id);
      IF id.scopeNm # NIL THEN f.WriteString(id.scopeNm) END;
      f.Write(keySy);
      f.Write4B(id.modKey);
      id.impOrd := f.iNxt; INC(f.iNxt);
    END;
  END EmitImportId;

(* ======================================= *)

  PROCEDURE (f : SymFile)EmitProcedureId(id : Id.PrcId),NEW;
  (*
  ** Procedure = prcSy Name FormalType.
  *)
  BEGIN
    f.Write(prcSy);
    f.WriteName(id);
    IF id.prcNm # NIL THEN f.WriteString(id.prcNm) END;
    f.FormalType(id.type(Ty.Procedure));
  END EmitProcedureId;

(* ======================================= *)

  PROCEDURE (f : SymFile)EmitMethodId(id : Id.MthId),NEW;
  (*
  ** Method = mthSy Name Byte Byte TypeOrd [strSy ] FormalType.
  *)
  BEGIN
    IF id.kind = Id.fwdMth THEN id := id.resolve(Id.MthId) END;
    f.Write(mthSy);
    f.WriteName(id);
    f.Write(ORD(id.mthAtt));
    f.Write(id.rcvFrm.parMod);
    f.EmitTypeOrd(id.rcvFrm.type);
    IF id.prcNm # NIL THEN f.WriteString(id.prcNm) END;
    f.FormalType(id.type(Ty.Procedure));
  END EmitMethodId;

(* ======================================= *)

  PROCEDURE moduleOrd(tpId : D.Idnt) : INTEGER;
    VAR impM : Id.BlkId;
  BEGIN
    IF  (tpId = NIL) OR 
	(tpId.dfScp = NIL) OR 
	(tpId.dfScp.kind = Id.modId) THEN
      RETURN 0;
    ELSE
      impM := tpId.dfScp(Id.BlkId);
      ASSERT(impM.impOrd # 0);
      RETURN impM.impOrd;
    END;
  END moduleOrd;

(* ======================================= *)

  PROCEDURE (f : SymFile)EmitTypeHeader(t : D.Type),NEW;
  (*
  **  TypeHeader = typSy Ord [fromS Ord Name].
  *)
    VAR mod : INTEGER;
	idt : D.Idnt;
  BEGIN
    WITH t : Ty.Record DO
      IF t.bindTp = NIL THEN 
	idt := t.idnt;
      ELSIF t.bindTp.dump = 0 THEN
	ASSERT(FALSE);
	idt := NIL;
(*
 *	t.bindTp.dump := f.oNxt; INC(f.oNxt);
 *	D.AppendType(f.work, t.bindTp);
 *	idt := t.bindTp.idnt;
 *)
      ELSE
	idt := t.bindTp.idnt;
      END;
    ELSE
      idt := t.idnt;
    END;
    mod := moduleOrd(t.idnt);
    f.Write(tDefS);
    f.WriteOrd(t.dump);
    IF mod # 0 THEN
      f.Write(fromS);
      f.WriteOrd(mod);
      f.WriteName(idt);
    END;
  END EmitTypeHeader;

(* ======================================= *)

  PROCEDURE (f : SymFile)EmitArrayType(t : Ty.Array),NEW;
  BEGIN
    f.EmitTypeHeader(t);
    IF t.force # D.noEmit THEN	(* Don't emit structure unless forced *)
      f.Write(arrSy);
      f.EmitTypeOrd(t.elemTp);
      IF t.length > 127 THEN
	f.Write(numSy);
	f.Write8B(t.length);
      ELSIF t.length > 0 THEN
	f.Write(bytSy);
	f.Write(t.length);
      END;
      f.Write(endAr);
    END;
  END EmitArrayType;

(* ======================================= *)

  PROCEDURE (f : SymFile)EmitRecordType(t : Ty.Record),NEW;
    VAR index  : INTEGER;
	field  : D.Idnt;
	method : D.Idnt;
	symVisit : SymFileSFA;
  (*
  **  Record = TypeHeader recSy recAtt [truSy | falSy | <others>] 
  **		[basSy TypeOrd] {Name TypeOrd} {Method} {Statics} endRc.
  *)
  BEGIN
    f.EmitTypeHeader(t);
    IF t.force # D.noEmit THEN	(* Don't emit structure unless forced *)
      f.Write(recSy);
      f.Write(t.recAtt);
   (* ########## *)
      IF t.recAtt = Ty.iFace THEN
  	f.Write(truSy);
      ELSIF G.special THEN  
	f.Write(falSy);
      END;
   (* ########## *)
      IF t.baseTp # NIL THEN
        f.Write(basSy);
	f.EmitTypeOrd(t.baseTp);
      END;
      FOR index := 0 TO t.fields.tide-1 DO
	field := t.fields.a[index];
	IF field.vMod # D.prvMode THEN
	  f.WriteName(field);
	  f.EmitTypeOrd(field.type);
	END;
      END;
      IF t.force = D.forced THEN  (* Don't emit methods unless forced *)
        FOR index := 0 TO t.methods.tide-1 DO
	  method := t.methods.a[index];
	  IF method.vMod # D.prvMode THEN
	    f.EmitMethodId(method(Id.MthId));
	  END;
        END;
	IF G.special THEN  (* we might need to emit static stuff *)
          FOR index := 0 TO t.statics.tide-1 DO
	    field := t.statics.a[index];
	    IF field.vMod # D.prvMode THEN
	      CASE field.kind OF
	      | Id.conId  : f.EmitConstId(field(Id.ConId));
	      | Id.varId  : f.EmitVariableId(field(Id.VarId));
	      | Id.conPrc : f.EmitProcedureId(field(Id.PrcId));
	      END;
	    END;
          END;
	END;
      END;
      f.Write(endRc);
    END;
    D.AppendType(f.modS.expRecs, t);
  END EmitRecordType;

(* ======================================= *)

  PROCEDURE (f : SymFile)EmitEnumType(t : Ty.Enum),NEW;
    VAR index  : INTEGER;
	const  : D.Idnt;
	symVisit : SymFileSFA;
  (*
  **  Enum = TypeHeader eTpSy { constant } endRc.
  *)
  BEGIN
    f.EmitTypeHeader(t);
    f.Write(eTpSy);
    FOR index := 0 TO t.statics.tide-1 DO
      const := t.statics.a[index];
      IF const.vMod # D.prvMode THEN f.EmitConstId(const(Id.ConId)) END;
    END;
    f.Write(endRc);
    (* D.AppendType(f.modS.expRecs, t); *)
  END EmitEnumType;

(* ======================================= *)

  PROCEDURE (f : SymFile)EmitOpaqueType(t : Ty.Opaque),NEW;
  BEGIN
    f.EmitTypeHeader(t);
  END EmitOpaqueType;

(* ======================================= *)

  PROCEDURE (f : SymFile)EmitPointerType(t : Ty.Pointer),NEW;
  BEGIN
    f.EmitTypeHeader(t);
    IF (t.force # D.noEmit) OR 			(* Only emit structure if *)
       (t.boundTp.force # D.noEmit) THEN	(* ptr or boundTp forced. *)
      f.Write(ptrSy);
      f.EmitTypeOrd(t.boundTp);
    END;
  END EmitPointerType;

(* ======================================= *)

  PROCEDURE (f : SymFile)EmitProcedureType(t : Ty.Procedure),NEW;
  BEGIN
  END EmitProcedureType;

(* ======================================= *)

  PROCEDURE (f : SymFile)EmitTypeList(),NEW;
    VAR indx : INTEGER;
	type : D.Type;
  BEGIN
   (*
    *   We cannot use a FOR loop here, as the tide changes
    *   during evaluation, as a result of reaching new types.
    *)
    indx := 0;
    WHILE indx < f.work.tide DO
      type := f.work.a[indx];
      WITH type : Ty.Array     DO f.EmitArrayType(type);
      |    type : Ty.Record    DO f.EmitRecordType(type);
      |    type : Ty.Opaque    DO f.EmitOpaqueType(type);
      |    type : Ty.Pointer   DO f.EmitPointerType(type);
      |    type : Ty.Procedure DO f.EmitProcedureType(type);
      |    type : Ty.Enum      DO f.EmitEnumType(type);
      END;
      INC(indx);
    END;
  END EmitTypeList;

(* ======================================= *)

  PROCEDURE EmitSymfile*(m : Id.BlkId);
    VAR fileName : FileNames.NameString;
	symVisit : SymFileSFA;
	symfile  : SymFile;
	marker   : INTEGER;
  (*
  ** SymFile = Header [String (falSy | truSy | <others>)]
  **		{Import | Constant | Variable
  **                 | Type | Procedure | Method} TypeList.
  ** Header = magic modSy Name.
  *)
  BEGIN
   (*
    *  Create the SymFile structure, and open the output file.
    *)
    symfile := newSymFile(m);
    D.getName.Of(m, fileName);
    fileName := fileName + ".cps";
    symfile.file := BF.createFile(fileName);
    IF symfile.file = NIL THEN
      S.SemError.Report(177, 0, 0);
      Error.WriteString("Cannot create file <" + fileName + ">"); 
      Error.WriteLn;
      RETURN;
    ELSE
     (*
      *  Emit the symbol file header
      *)
      IF G.verbose THEN G.Message("Created " + fileName) END;
      IF D.rtsMd IN m.xAttr THEN
	marker := RTS.loInt(syMag);	(* ==> a system module *)
      ELSE
	marker := RTS.loInt(magic);	(* ==> a normal module *)
      END;
      symfile.Write4B(RTS.loInt(marker));
      symfile.Write(modSy);
      symfile.WriteName(m);
      IF m.scopeNm # NIL THEN (* explicit name *)
	symfile.WriteString(m.scopeNm);
	symfile.Write(falSy);
      END;
     (*
      *  Create the symbol table visitor, an extension of 
      *  Symbols.SymForAll type.  Emit symbols from the scope.
      *)
      NEW(symVisit);
      symVisit.sym := symfile;
      symfile.modS.symTb.Apply(symVisit); 
     (*
      *  Now emit the types on the worklist.
      *)
      symfile.Write(start);
      symfile.EmitTypeList();
      symfile.Write(close);
     (*
      *  Now emit the accumulated checksum key symbol.
      *)
      symfile.Write(keySy);
      lastKeyVal := symfile.cSum;
      symfile.Write4B(symfile.cSum);
      BF.CloseFile(symfile.file);
    END;
  END EmitSymfile;

(* ============================================================ *)
(* ========	Various reading utility procedures	======= *)
(* ============================================================ *)

  PROCEDURE read(f : BF.FILE) : INTEGER;
  BEGIN
    RETURN BF.readByte(f);
  END read;

(* ======================================= *)

  PROCEDURE ReadUTF(f : BF.FILE; OUT nam : ARRAY OF CHAR);
    CONST
	bad = "Bad UTF-8 string";
    VAR num : INTEGER;
	bNm : INTEGER;
	idx : INTEGER;
	chr : INTEGER;
  BEGIN
    num := 0;
    bNm := read(f) * 256 + read(f);
    FOR idx := 0 TO bNm-1 DO
      chr := read(f);
      IF chr <= 07FH THEN
	nam[num] := CHR(chr); INC(num);
      ELSIF chr DIV 32 = 06H THEN
	bNm := chr MOD 32 * 64;
	chr := read(f);
	IF chr DIV 64 = 02H THEN
	  nam[num] := CHR(bNm + chr DIV 64); INC(num);
	ELSE
	  RTS.Throw(bad);
	END;
      ELSIF chr DIV 16 = 0EH THEN
	bNm := chr MOD 16 * 64;
	chr := read(f);
	IF chr DIV 64 = 02H THEN
	  bNm := bNm + chr DIV 64; 
	  chr := read(f);
	  IF chr DIV 64 = 02H THEN
	    nam[num] := CHR(bNm + chr DIV 64); INC(num);
	  ELSE 
	    RTS.Throw(bad);
	  END;
	ELSE
	  RTS.Throw(bad);
	END;
      ELSE
	RTS.Throw(bad);
      END;
    END;
    nam[num] := 0X;
  END ReadUTF;

(* ======================================= *)

  PROCEDURE readChar(f : BF.FILE) : CHAR;
  BEGIN
    RETURN CHR(read(f) * 256 + read(f));
  END readChar;

(* ======================================= *)

  PROCEDURE readInt(f : BF.FILE) : INTEGER;
  BEGIN [UNCHECKED_ARITHMETIC]
    (* overflow checking off here *)
    RETURN ((read(f) * 256 + read(f)) * 256 + read(f)) * 256 + read(f);
  END readInt;

(* ======================================= *)

  PROCEDURE readLong(f : BF.FILE) : LONGINT;
    VAR result : LONGINT;
	index  : INTEGER;
  BEGIN [UNCHECKED_ARITHMETIC]
    (* overflow checking off here *)
    result := read(f);
    FOR index := 1 TO 7 DO
      result := result * 256 + read(f);
    END;
    RETURN result;
  END readLong;

(* ======================================= *)

  PROCEDURE readReal(f : BF.FILE) : REAL;
    VAR result : LONGINT;
  BEGIN
    result := readLong(f);
    RETURN RTS.longBitsToReal(result);
  END readReal;

(* ======================================= *)

  PROCEDURE readOrd(f : BF.FILE) : INTEGER;
    VAR chr : INTEGER;
  BEGIN
    chr := read(f);
    IF chr <= 07FH THEN RETURN chr;
    ELSE
      DEC(chr, 128);
      RETURN chr + read(f) * 128;
    END;
  END readOrd;

(* ============================================================ *)
(* ========		Symbol File Reader		======= *)
(* ============================================================ *)

  PROCEDURE newSymFileReader*(mod : Id.BlkId) : SymFileReader;
    VAR new : SymFileReader;
  BEGIN
    NEW(new);
    new.modS := mod;
    D.InitTypeSeq(new.tArray, 8);
    D.InitScpSeq(new.sArray, 8);
    RETURN new;
  END newSymFileReader;

(* ======================================= *)
  PROCEDURE^ (f : SymFileReader)SymFile(IN nm : ARRAY OF CHAR),NEW;
  PROCEDURE^ WalkThisImport(imp, mod : Id.BlkId);
(* ======================================= *)

  PROCEDURE Abandon(f : SymFileReader);
  BEGIN
    RTS.Throw("Bad symbol file format" + 
	      Nh.charOpenOfHash(f.impS.hash)^); 
  END Abandon;

(* ======================================= *)

  PROCEDURE (f : SymFileReader)GetSym(),NEW;
    VAR file : BF.FILE;
  BEGIN
    file := f.file;
    f.sSym := read(file);
    CASE f.sSym OF
    | namSy : 
	f.iAtt := read(file); ReadUTF(file, f.sAtt);
    | strSy : 
	ReadUTF(file, f.sAtt);
    | retSy, fromS, tDefS, basSy :
	f.iAtt := readOrd(file);
    | bytSy :
	f.iAtt := read(file);
    | keySy, setSy :
	f.iAtt := readInt(file);
    | numSy :
	f.lAtt := readLong(file);
    | fltSy :
	f.rAtt := readReal(file);
    | chrSy :
	f.cAtt := readChar(file);
    ELSE (* nothing to do *)
    END;
  END GetSym;

(* ======================================= *)

  PROCEDURE (f : SymFileReader)ReadPast(sym : INTEGER),NEW;
  BEGIN
    IF f.sSym # sym THEN Abandon(f) END;
    f.GetSym();
  END ReadPast;

(* ======================================= *)

  PROCEDURE (f : SymFileReader)Parse*(scope : Id.BlkId;
				   IN filNm : FileNames.NameString),NEW;
    VAR fileName : FileNames.NameString;
	marker   : INTEGER;
  BEGIN
    f.impS := scope;
    D.AppendScope(f.sArray, scope);
    fileName := filNm + ".cps";
    f.file := BF.findOnPath("CPSYM", fileName);
    IF f.file = NIL THEN
      IF scope.token = NIL THEN
	Error.WriteString("File <" + fileName + "> not found"); 
	Error.WriteLn;
      ELSE
	S.SemError.Report(129, scope.token.lin, scope.token.col);
      END;
      RETURN;
    ELSE
      IF G.verbose THEN G.Message("Opened " + fileName) END;
      marker := readInt(f.file);
      IF marker = RTS.loInt(magic) THEN
	(* normal case, nothing to do *)
      ELSIF marker = RTS.loInt(syMag) THEN
        INCL(scope.xAttr, D.rtsMd);
      ELSE
	IF scope.token = NIL THEN
	  Error.WriteString("File <" + fileName + "> wrong format"); 
	  Error.WriteLn;
	ELSE
          S.SemError.Report(130, scope.token.lin, scope.token.col);
	END;
        RETURN;
      END;
      f.GetSym();
      f.SymFile(filNm);
      IF G.verbose THEN 
	G.Message("Ended " + fileName + ", Key: " 
			+ LitValue.intToCharOpen(f.impS.modKey)^);
      END;
      BF.CloseFile(f.file);
    END;
  END Parse;

(* ============================================ *)

  PROCEDURE testInsert(id : D.Idnt; sc : D.Scope) : D.Idnt;
    VAR ident : D.Idnt;

    PROCEDURE Report(i,s : D.Idnt);
      VAR iS, sS : FileNames.NameString;
    BEGIN
      D.getName.Of(i, iS);
      D.getName.Of(s, sS);
      S.SemError.RepSt2(172, iS, sS, S.line, S.col);
    END Report;

  BEGIN
    IF sc.symTb.enter(id.hash, id) THEN
      ident := id;
    ELSE
      ident := sc.symTb.lookup(id.hash);	(* Warp the return Idnt	*)
      IF ident.kind # id.kind THEN Report(id, sc); ident := id END;
    END;
    RETURN ident;
  END testInsert;

(* ============================================ *)

  PROCEDURE Insert(id : D.Idnt; VAR tb : D.SymbolTable);

    PROCEDURE Report(i : D.Idnt);
      VAR iS : FileNames.NameString;
    BEGIN
      D.getName.Of(i, iS);
      S.SemError.RepSt1(172, iS, 1, 1);
    END Report;

  BEGIN
    IF ~tb.enter(id.hash, id) THEN Report(id) END;
  END Insert;

(* ============================================ *)

  PROCEDURE (f : SymFileReader)getLiteral() : D.Expr,NEW;
    VAR expr : D.Expr;
  BEGIN
    CASE f.sSym OF
    | truSy : expr := ExprDesc.mkTrueX();
    | falSy : expr := ExprDesc.mkFalseX();
    | numSy : expr := ExprDesc.mkNumLt(f.lAtt);
    | chrSy : expr := ExprDesc.mkCharLt(f.cAtt);
    | fltSy : expr := ExprDesc.mkRealLt(f.rAtt);
    | setSy : expr := ExprDesc.mkSetLt(BITS(f.iAtt));
    | strSy : expr := ExprDesc.mkStrLt(f.sAtt);		(* implicit f.sAtt^ *)
    END;
    f.GetSym();						(* read past value  *)
    RETURN expr;
  END getLiteral;

(* ============================================ *)

  PROCEDURE (f : SymFileReader)typeOf(ord : INTEGER) : D.Type,NEW;
    VAR newT : D.Type;
	indx : INTEGER;
  BEGIN
    IF ord < D.tOffset THEN				(* builtin type	*)	
      RETURN B.baseTypeArray[ord];
    ELSIF ord - D.tOffset < f.tArray.tide THEN
      RETURN f.tArray.a[ord - D.tOffset];
    ELSE 
      indx := f.tArray.tide + D.tOffset;
      REPEAT
        newT := Ty.newTmpTp();
        newT.dump := indx; INC(indx);
        D.AppendType(f.tArray, newT);
      UNTIL indx > ord;
      RETURN newT;
    END;
  END typeOf;

(* ============================================ *)

  PROCEDURE (f : SymFileReader)getTypeFromOrd() : D.Type,NEW;
    VAR ord : INTEGER;
  BEGIN
    ord := readOrd(f.file);
    f.GetSym();
    RETURN f.typeOf(ord);
  END getTypeFromOrd;

(* ============================================ *)

  PROCEDURE (f : SymFileReader)getFormalType(i : INTEGER) : D.Type,NEW;
  (*
  // FormalType = [retSy TypeOrd] frmSy {parSy byte TypeOrd} endFm.
  //	-- optional phrase is return type for proper procedures
  *)
    VAR rslt : Ty.Procedure;
 	parD : Id.ParId;
	byte : INTEGER;
  BEGIN
    rslt := Ty.newPrcTp();
    IF f.sSym = retSy THEN 
      rslt.retType := f.typeOf(f.iAtt);
      f.GetSym();
    END;
    f.ReadPast(frmSy);
    WHILE f.sSym = parSy DO
      byte := read(f.file);
      parD := Id.newParId();
      parD.parMod := byte;
      parD.varOrd := i; 
      parD.type := f.getTypeFromOrd();
      Id.AppendParam(rslt.formals, parD);
      INC(i);
    END;
    f.ReadPast(endFm);
    RETURN rslt;
  END getFormalType;

(* ============================================ *)

  PROCEDURE (f : SymFileReader)pointerType(old : D.Type) : D.Type,NEW;
  (* Assert: the current symbol is ptrSy.	*)
  (* Pointer   = TypeHeader ptrSy TypeOrd.	*)
    VAR rslt : Ty.Pointer;
	indx : INTEGER;
	junk : D.Type;
  BEGIN
    indx := readOrd(f.file);
    WITH old : Ty.Pointer DO
      rslt := old;
     (*
      *  Check if there is space in the tArray for this
      *  element, otherwise expand using typeOf().
      *)
      IF indx - D.tOffset >= f.tArray.tide THEN
	junk := f.typeOf(indx);
      END;
      f.tArray.a[indx - D.tOffset] := rslt.boundTp;
    ELSE
      rslt := Ty.newPtrTp();
      rslt.boundTp := f.typeOf(indx);
    END;
    f.GetSym();
    RETURN rslt;
  END pointerType;

(* ============================================ *)

  PROCEDURE (f : SymFileReader)procedureType() : D.Type,NEW;
  (* Assert: the current symbol is prcSy.	*)
  (* ProcType  = TypeHeader pTpSy FormalType.	*)
  BEGIN
    RETURN f.getFormalType(0);
  END procedureType;

(* ============================================ *)

  PROCEDURE (f : SymFileReader)arrayType() : D.Type,NEW;
  (* Assert: at entry the current symbol is arrSy.		     *)
  (* Array      = TypeHeader arrSy TypeOrd (Byte | Number | ) endAr. *)
  (*	-- nullable phrase is array length for fixed length arrays   *)
    VAR rslt : Ty.Array;
	eTyp : D.Type;
  BEGIN
    rslt := Ty.newArrTp();
    rslt.elemTp := f.typeOf(readOrd(f.file));
    f.GetSym();
    IF f.sSym = bytSy THEN
      rslt.length := f.iAtt;
      f.GetSym();
    ELSIF f.sSym = numSy THEN
      rslt.length := SHORT(f.lAtt);
      f.GetSym();
    (* ELSE length := 0 *)
    END;
    f.ReadPast(endAr);
    RETURN rslt;
  END arrayType;

(* ============================================ *)
  PROCEDURE^ (f : SymFileReader)procedure() : Id.PrcId,NEW;
  PROCEDURE^ (f : SymFileReader)method()    : Id.MthId,NEW;
  PROCEDURE^ (f : SymFileReader)constant()  : Id.ConId,NEW;
  PROCEDURE^ (f : SymFileReader)variable()  : Id.VarId,NEW;
(* ============================================ *)

  PROCEDURE (f : SymFileReader)recordType(old : D.Type) : D.Type,NEW;
  (* Assert: at entry the current symbol is recSy.			*)
  (* Record     = TypeHeader recSy recAtt [truSy | falSy | <others>] 	*)
  (*	[basSy TypeOrd] {Name TypeOrd} {Method} {Statics} endRc.	*)
    VAR rslt : Ty.Record;
	fldD : Id.FldId;
	varD : Id.VarId;
	mthD : Id.MthId;
	prcD : Id.PrcId;
	oldS : INTEGER;
  BEGIN
    WITH old : Ty.Record DO rslt := old ELSE rslt := Ty.newRecTp() END;
    rslt.recAtt := read(f.file);
    IF rslt.recAtt # Ty.noAtt THEN INCL(rslt.xAttr, D.clsTp) END;

    f.GetSym();				(* Get past recSy rAtt	*)
    IF f.sSym = falSy THEN
      INCL(rslt.xAttr, D.isJv);
      f.GetSym();
    ELSIF f.sSym = truSy THEN
      INCL(rslt.xAttr, D.isJv);
      INCL(rslt.xAttr, D.jvInf);
      f.GetSym();
    END;
    IF f.impS.scopeNm # NIL THEN rslt.extrnNm := f.impS.scopeNm END;

    IF f.sSym = basSy THEN
      rslt.baseTp := f.typeOf(f.iAtt);
      IF f.iAtt # Ty.anyRec THEN INCL(rslt.xAttr, D.clsTp) END;
      f.GetSym();
    END;
    WHILE f.sSym = namSy DO
      fldD := Id.newFldId();
      fldD.SetMode(f.iAtt);
      fldD.hash := Nh.enterStr(f.sAtt);
      fldD.type := f.typeOf(readOrd(f.file));
      fldD.recTyp := rslt;
      f.GetSym();
      IF rslt.symTb.enter(fldD.hash, fldD) THEN 
	D.AppendIdnt(rslt.fields, fldD);
      END;
    END;

    WHILE (f.sSym = mthSy) OR
	  (f.sSym = prcSy) OR
	  (f.sSym = varSy) OR
	  (f.sSym = conSy) DO
      oldS := f.sSym; f.GetSym();
      IF oldS = mthSy THEN
	mthD := f.method();
        mthD.bndType := rslt;
        mthD.type(Ty.Procedure).receiver := rslt;
        IF rslt.symTb.enter(mthD.hash, mthD) THEN 
	  D.AppendIdnt(rslt.methods, mthD);
        END;
      ELSIF oldS = prcSy THEN
	prcD := f.procedure();
        prcD.bndType := rslt;
	Insert(prcD, rslt.symTb);
	D.AppendIdnt(rslt.statics, prcD);
      ELSIF oldS = varSy THEN
	varD := f.variable();
	varD.recTyp := rslt;
	Insert(varD, rslt.symTb);
	D.AppendIdnt(rslt.statics, varD);
      ELSIF oldS = conSy THEN
	Insert(f.constant(), rslt.symTb);
      ELSE
	Abandon(f);
      END;
    END;
    f.ReadPast(endRc); 
    RETURN rslt;
  END recordType;

(* ============================================ *)

  PROCEDURE (f : SymFileReader)enumType() : D.Type,NEW;
  (* Assert: at entry the current symbol is eTpSy.			*)
  (* Enum  = TypeHeader eTpSy { Constant} endRc.			*)
    VAR rslt : Ty.Enum;
  BEGIN
    rslt := Ty.newEnuTp();
    f.GetSym();				(* Get past recSy 	*)
    WHILE f.sSym = conSy DO
      f.GetSym();
      Insert(f.constant(), rslt.symTb);
    END;
    f.ReadPast(endRc); 
    RETURN rslt;
  END enumType;

(* ============================================ *)

  PROCEDURE (f : SymFileReader)Type(),NEW;
  (* Type       = typSy Name TypeOrd.		*)
    VAR newI : Id.TypId;
	oldI : D.Idnt;
	type : D.Type;
  BEGIN
   (* 
    * Post: every previously unknown typId id
    *	has the property:  id.type.idnt = id.
    *   If oldI # newT, then the new typId has
    *   newT.type.idnt = oldI.
    *)
    newI := Id.newTypId(NIL);
    newI.SetMode(f.iAtt);
    newI.hash := Nh.enterStr(f.sAtt);
    newI.type := f.getTypeFromOrd(); 
    newI.dfScp := f.impS;
    oldI := testInsert(newI, f.impS);

    IF oldI # newI THEN 
      f.tArray.a[newI.type.dump - D.tOffset] := oldI.type;
    END;

    IF newI.type.idnt = NIL THEN newI.type.idnt := oldI END;
  END Type;

(* ============================================ *)

  PROCEDURE (f : SymFileReader)Import(),NEW;
  (* Import     = impSy Name [String] Key.	*)
  (*	-- optional string is external name	*)
  (* first symbol should be namSy here.		*)
    VAR impD : Id.BlkId;
	oldS : Id.BlkId;
	oldD : D.Idnt;
  BEGIN
    impD := Id.newImpId();
    impD.dfScp := impD;			(* ImpId define their own scope *)

    INCL(impD.xAttr, D.weak);
    impD.SetMode(f.iAtt);
    impD.hash := Nh.enterStr(f.sAtt);
    f.ReadPast(namSy); 
    IF impD.hash = f.modS.hash THEN	(* Importing own imp indirectly	*)
					(* Shouldn't this be an error?  *)
      D.AppendScope(f.sArray, f.modS);
      IF f.sSym = strSy THEN 
	(* probably don't need to do anything here ... *)
	f.GetSym();
      END;
    ELSE				(* Importing some other module.	*)
      oldD := testInsert(impD, f.modS);
      IF f.sSym = strSy THEN 
	impD.scopeNm := LitValue.strToCharOpen(f.sAtt);
        f.GetSym();
      END;
      IF (oldD # impD) & (oldD.kind = Id.impId) THEN
	oldS := oldD(Id.BlkId);
        D.AppendScope(f.sArray, oldS);
	IF (oldS.modKey # 0) & (f.iAtt # oldS.modKey) THEN
	  S.SemError.RepSt1(133,		(* Detected bad KeyVal	*)
		Nh.charOpenOfHash(impD.hash)^, 
		S.line, S.col);
	END;
        IF ~(D.weak IN oldS.xAttr) &
	   ~(D.fixd IN oldS.xAttr) THEN
	 (*
	  *   This recursively reads the symbol files for 
	  *   any imports of this file which are on the
	  *   list to be imported later anyhow.
	  *)
	  WalkThisImport(oldS, f.modS);
        END;
      ELSE
        D.AppendScope(f.sArray, impD);
      END;
      impD.modKey := f.iAtt;
    END;
    f.ReadPast(keySy);
  END Import;

(* ============================================ *)
(*
 *PROCEDURE (f : SymFileReader)Constant(),NEW;
 *(* Constant   = conSy Name Literal.		*)
 *(* Assert: f.sSym = namSy.			*)
 *  VAR newC : Id.ConId;
 *	anyI : D.Idnt;
 *BEGIN
 *  newC := Id.newConId();
 *  newC.SetMode(f.iAtt);
 *  newC.hash := Nh.enterStr(f.sAtt);
 *  newC.dfScp := f.impS;
 *  f.ReadPast(namSy);
 *  newC.conExp := f.getLiteral();
 *  newC.type := newC.conExp.type;
 *  anyI := testInsert(newC, f.impS);
 *END Constant;
 *)
(* ============================================ *)

  PROCEDURE (f : SymFileReader)constant() : Id.ConId,NEW;
  (* Constant   = conSy Name Literal.		*)
  (* Assert: f.sSym = namSy.			*)
    VAR newC : Id.ConId;
	anyI : D.Idnt;
  BEGIN
    newC := Id.newConId();
    newC.SetMode(f.iAtt);
    newC.hash := Nh.enterStr(f.sAtt);
    newC.dfScp := f.impS;
    f.ReadPast(namSy);
    newC.conExp := f.getLiteral();
    newC.type := newC.conExp.type;
    RETURN newC;
  END constant;

(* ============================================ *)
(*
 *PROCEDURE (f : SymFileReader)Variable(),NEW;
 *(* Variable   = varSy Name TypeOrd.		*)
 *  VAR newV : Id.VarId;
 *	anyI : D.Idnt;
 *BEGIN
 *  newV := Id.newVarId();
 *  newV.SetMode(f.iAtt);
 *  newV.hash := Nh.enterStr(f.sAtt);
 *  newV.type := f.getTypeFromOrd();
 *  newV.dfScp := f.impS;
 *  anyI := testInsert(newV, f.impS);
 *END Variable;
 *)
(* ============================================ *)

  PROCEDURE (f : SymFileReader)variable() : Id.VarId,NEW;
  (* Variable   = varSy Name TypeOrd.		*)
    VAR newV : Id.VarId;
	anyI : D.Idnt;
  BEGIN
    newV := Id.newVarId();
    newV.SetMode(f.iAtt);
    newV.hash := Nh.enterStr(f.sAtt);
    newV.type := f.getTypeFromOrd();
    newV.dfScp := f.impS;
    RETURN newV;
  END variable;

(* ============================================ *)
(*
 *PROCEDURE (f : SymFileReader)Procedure(),NEW;
 *(* Procedure  = prcSy Name[String]FormalType. *)
 *(* This is a static proc, mths come with Recs *)
 *  VAR newP : Id.PrcId;
 *	anyI : D.Idnt;
 *BEGIN
 *  newP := Id.newPrcId();
 *  newP.setPrcKind(Id.conPrc);
 *  newP.SetMode(f.iAtt);
 *  newP.hash := Nh.enterStr(f.sAtt);
 *  newP.dfScp := f.impS;
 *  f.ReadPast(namSy);
 *  IF f.sSym = strSy THEN 
 *    newP.prcNm := LitValue.strToCharOpen(f.sAtt);
 *   (* and leave scopeNm = NIL *)
 *    f.GetSym();
 *  END;
 *  newP.type := f.getFormalType(0);
 *  (* IF this is a java module, do some semantic checks *)
 *  (* ... *)
 *  anyI := testInsert(newP, f.impS);
 *END Procedure;
 *)
(* ============================================ *)

  PROCEDURE (f : SymFileReader)procedure() : Id.PrcId,NEW;
  (* Procedure  = prcSy Name[String]FormalType. *)
  (* This is a static proc, mths come with Recs *)
    VAR newP : Id.PrcId;
	anyI : D.Idnt;
  BEGIN
    newP := Id.newPrcId();
    newP.setPrcKind(Id.conPrc);
    newP.SetMode(f.iAtt);
    newP.hash := Nh.enterStr(f.sAtt);
    newP.dfScp := f.impS;
    f.ReadPast(namSy);
    IF f.sSym = strSy THEN 
      newP.prcNm := LitValue.strToCharOpen(f.sAtt);
     (* and leave scopeNm = NIL *)
      f.GetSym();
    END;
    newP.type := f.getFormalType(0);
    (* IF this is a java module, do some semantic checks *)
    (* ... *)
    RETURN newP;
  END procedure;

(* ============================================ *)

  PROCEDURE (f : SymFileReader)method() : Id.MthId,NEW;
  (* Method     = mthSy Name byte byte TypeOrd [String] FormalType. *)
    VAR newM : Id.MthId;
	rcvD : Id.ParId;
	rFrm : INTEGER;
	mAtt : SET;
  BEGIN
    newM := Id.newMthId();
    newM.SetMode(f.iAtt);
    newM.setPrcKind(Id.conMth);
    newM.hash := Nh.enterStr(f.sAtt);
    newM.dfScp := f.impS;
    rcvD := Id.newParId();
    rcvD.varOrd := 0;
   (* byte1 is the method attributes  *)
    mAtt := BITS(read(f.file));
   (* byte2 is param form of receiver *)
    rFrm := read(f.file);
   (* next 1 or 2 bytes are rcv-type  *)
    rcvD.type := f.typeOf(readOrd(f.file));
    f.GetSym();
    rcvD.parMod := rFrm;
    IF f.sSym = strSy THEN 
      newM.prcNm := LitValue.strToCharOpen(f.sAtt);
     (* and leave scopeNm = NIL *)
      f.GetSym();
    END;
    newM.type := f.getFormalType(1);
    newM.mthAtt := mAtt;
    newM.rcvFrm := rcvD;
   (* IF this is a java module, do some semantic checks *)
    RETURN newM;
  END method;

(* ============================================ *)

  PROCEDURE (f : SymFileReader)TypeList(),NEW;
  (* TypeList   = start { Array | Record | Pointer 	*)
  (*		  | ProcType } close. 			*)
  (* TypeHeader = tDefS Ord [fromS Ord Name].		*)
    VAR modOrd : INTEGER;
	typOrd : INTEGER;
	typIdx : INTEGER;
	tpDesc : D.Type;
	tpTemp : D.Type;
	tpIdnt : Id.TypId;
	prevId : D.Idnt;
	impScp : D.Scope;
	linkIx : INTEGER;
	bndTyp : D.Type;
	typeFA : TypeLinker;
  BEGIN
    WHILE f.sSym = tDefS DO
      tpIdnt := NIL;
     (* Do type header *)
      typOrd := f.iAtt;
      typIdx := typOrd - D.tOffset;
      tpTemp := f.tArray.a[typIdx];
      f.ReadPast(tDefS);
     (*
      *  The fromS symbol appears if the type is imported.
      *)
      IF f.sSym = fromS THEN
	modOrd := f.iAtt;
	impScp := f.sArray.a[modOrd];
	f.GetSym();
(*
 *   With the strict ordering of the imports,
 *   it may be unnecessary to create this object
 *   in case the other module has been fully read
 *   already?
 *   	It is also possible that the type has 
 *   been imported already, but just as an opaque.
 *)
	tpIdnt := Id.newTypId(NIL);
	tpIdnt.SetMode(f.iAtt);
	tpIdnt.hash := Nh.enterStr(f.sAtt);
	tpIdnt.dfScp := impScp;
	tpIdnt := testInsert(tpIdnt, impScp)(Id.TypId);
	f.ReadPast(namSy);
      END;
     (* Get type info. *)
      CASE f.sSym OF
      | arrSy : tpDesc := f.arrayType();
      | recSy : tpDesc := f.recordType(tpTemp);
      | ptrSy : tpDesc := f.pointerType(tpTemp);
      | pTpSy : tpDesc := f.procedureType();
      | eTpSy : tpDesc := f.enumType();
      ELSE 
	tpDesc := Ty.newNamTp();
      END;
      IF tpIdnt # NIL THEN
       (*
	*  A name has been declared for this type, tpIdnt is
	*  the (possibly previously known) id descriptor, and
	*  tpDesc is the newly parsed descriptor of the type.
	*  
	*  Because imports are partially ordered, if a 
	*  (non-opaque) type descriptor already exists, 
	*  it is definitive. We therefore override
	*  the newly parsed with the old.
	*)
	IF  tpIdnt.type = NIL THEN
	  tpIdnt.type := tpDesc;
	  tpDesc.idnt := tpIdnt;
	ELSIF ~(tpDesc IS Ty.Opaque) &
	       (tpIdnt.type IS Ty.Opaque) THEN
	  tpIdnt.type(Ty.Opaque).resolved := tpDesc;
	  tpIdnt.type := tpDesc;
	  tpDesc.idnt := tpIdnt;
	ELSE 
	 (*
	  *  Normally, imported types cannot be anonymous.
	  *  However, there is one special case here. Anon
	  *  records can be base types, but are always 
	  *  preceeded by the binding pointer type. Thus
	  *  we pre-override the future record declaration
	  *  by the bound type of the pointer.  This ensures
	  *  uniqueness of the record descriptor.
	  *)
	  WITH tpDesc : Ty.Pointer DO
	    bndTyp := tpDesc.boundTp;
	    IF bndTyp.kind = Ty.tmpTp THEN
	      linkIx := bndTyp.dump - D.tOffset;
	      IF (linkIx > typIdx) &
	         (tpIdnt.type IS Ty.Pointer) THEN 
		f.tArray.a[linkIx] := tpIdnt.type(Ty.Pointer).boundTp;
	      END;
	    END;
	  ELSE (* skip *)
	  END;
	 (* 
	  *  Now throw away the newly parsed type desc, 
	  *  and use the previously known definitive type.
	  *)
	  tpDesc := tpIdnt.type;
	END;
        f.tArray.a[typIdx] := tpDesc;
      ELSE  
       (* 
	*  tpIdnt is NIL ==> type is from this import,
	*  except for the special case above.  In the usual
	*  case we replace the tmpTp by tpDesc. In the special
	*  case the tmpTp has been already been overridden by 
	*  the previously imported bound type.
	*)
	IF f.tArray.a[typIdx].dump = typOrd THEN
	  (* the usual case ... *)
          prevId := f.tArray.a[typIdx].idnt;
          IF (prevId # NIL) &
             (prevId.type.kind = Ty.namTp)  THEN
            prevId.type(Ty.Opaque).resolved := tpDesc;
            prevId.type := tpDesc;
	  END;
	  tpDesc.idnt := prevId;
          f.tArray.a[typIdx] := tpDesc;
	  (* ELSE nothing to do *)
	END;
      END;
    END; (* while *)
   (*
    *  First we fix up all symbolic references in the
    *  the type array.  Postcondition is : no element
    *  of the type array directly or indirectly refers
    *  to a temporary type.
    *)
    FOR linkIx := 0 TO f.tArray.tide - 1 DO
      f.tArray.a[linkIx].TypeFix(f.tArray);
    END;
   (*
    *  We now fix up all references in the symbol table
    *  that still refer to temporary symbol-file types.
    *)
    NEW(typeFA);
    typeFA.sym := f;
    f.impS.symTb.Apply(typeFA); 
    f.ReadPast(close);
  END TypeList;

(* ============================================ *)

  PROCEDURE (f : SymFileReader)SymFile(IN nm : ARRAY OF CHAR),NEW;
   (*
   // SymFile    = Header [String (falSy | truSy | <others>)]
   //		{Import | Constant | Variable | Type | Procedure} 
   //		TypeList Key.
   // Header     = magic modSy Name.
   //
   //  magic has already been recognized.
   *)
    VAR oldS : INTEGER;
idnt : D.Idnt;
  BEGIN
    f.ReadPast(modSy);
    IF f.sSym = namSy THEN (* do something with f.sAtt *)
      IF nm # f.sAtt THEN
        Error.WriteString("Wrong name in symbol file. Expected <");
        Error.WriteString(nm + ">, found <");
        Error.WriteString(f.sAtt + ">"); 
	Error.WriteLn;
        HALT(1);
      END;
      f.GetSym();
    ELSE RTS.Throw("Bad symfile header");
    END;
    IF f.sSym = strSy THEN (* optional name *)
      f.impS.scopeNm := LitValue.strToCharOpen(f.sAtt);
      f.GetSym();
      IF f.sSym = falSy THEN 
        INCL(f.impS.xAttr, D.isJv);
	f.GetSym();
      ELSIF f.sSym = truSy THEN
        INCL(f.impS.xAttr, D.isJv);
        INCL(f.impS.xAttr, D.jvInf);
	f.GetSym();
      ELSE RTS.Throw("Bad explicit name");
      END; 
    END; 
    LOOP
      oldS := f.sSym;
      f.GetSym();
      CASE oldS OF
      | start : EXIT;
      | typSy : f.Type();
      | impSy : f.Import();
(*
 *    | conSy : f.Constant();
 *    | varSy : f.Variable();
 *    | prcSy : f.Procedure();
 *)
      | conSy : Insert(f.constant(),  f.impS.symTb);
      | varSy : Insert(f.variable(),  f.impS.symTb);
      | prcSy : Insert(f.procedure(), f.impS.symTb);
      ELSE RTS.Throw("Bad object");
      END;
    END;
    f.TypeList();
    IF f.sSym = keySy THEN
      IF f.impS.modKey = 0 THEN 
	f.impS.modKey := f.iAtt;
      ELSIF f.impS.modKey # f.iAtt THEN
	S.SemError.Report(173, S.line, S.col);	(* Detected bad KeyVal	*)
      END;
    ELSE RTS.Throw("Missing keySy");
    END; 
  END SymFile;

(* ============================================================ *)
(* ========	     SymFileSFA visitor method		======= *)
(* ============================================================ *)

  PROCEDURE (t : SymFileSFA)Op*(id : D.Idnt);
  BEGIN
    IF (id.kind = Id.impId) OR (id.vMod # D.prvMode) THEN
      CASE id.kind OF
      | Id.typId  : t.sym.EmitTypeId(id(Id.TypId));
      | Id.conId  : t.sym.EmitConstId(id(Id.ConId));
      | Id.impId  : t.sym.EmitImportId(id(Id.BlkId));
      | Id.varId  : t.sym.EmitVariableId(id(Id.VarId));
      | Id.conPrc : t.sym.EmitProcedureId(id(Id.PrcId));
      ELSE (* skip *)
      END;
    END;
  END Op;

(* ============================================================ *)
(* ========	     TypeLinker visitor method		======= *)
(* ============================================================ *)

  PROCEDURE (t : TypeLinker)Op*(id : D.Idnt);
  BEGIN
    IF id.type = NIL THEN RETURN
    ELSIF id.type.kind = Ty.tmpTp THEN
      id.type := Ty.update(t.sym.tArray, id.type);
    ELSE
      id.type.TypeFix(t.sym.tArray);
    END;
    IF  (id IS Id.TypId) & 
	(id.type.idnt = NIL) THEN id.type.idnt := id END;
  END Op;

(* ============================================================ *)
(* ========	    Symbol file parser method		======= *)
(* ============================================================ *)

  PROCEDURE WalkThisImport(imp, mod : Id.BlkId);
    VAR syFil : SymFileReader;
	filNm : FileNames.NameString;
  BEGIN
    PushStack(imp);
(*
IF G.verbose THEN
imp.Diagnose(0);
Console.WriteLn;
END;
 *)
    INCL(imp.xAttr, D.fixd);
    S.GetString(imp.token.pos, imp.token.len, filNm);
    syFil := newSymFileReader(mod);
    syFil.Parse(imp, filNm);
(*
IF G.verbose THEN
imp.Diagnose(0);
Console.WriteLn;
END;
 *)
    PopStack;
  END WalkThisImport;
 
(* ============================================ *)

  PROCEDURE WalkImports*(IN imps : D.ScpSeq; modI : Id.BlkId);
    VAR indx : INTEGER;
	scpI : D.Scope;
	blkI : Id.BlkId;
  BEGIN
    InitStack;
    FOR indx := 0 TO imps.tide-1 DO
      scpI := imps.a[indx];
      blkI := scpI(Id.BlkId);
      IF blkI.kind = Id.alias THEN
	blkI.symTb  := blkI.dfScp.symTb;
      ELSIF ~(D.fixd IN blkI.xAttr) THEN 
	WalkThisImport(blkI,modI);
      END;
    END;
  END WalkImports;

(* ============================================================ *)
BEGIN
  lastKeyVal := 0;
END SymFileRW.
(* ============================================================ *)
